#/***************************************************************************
# *   Copyright (c) 2016 Victor Titov (DeepSOIC) <vv.titov@gmail.com>       *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This library is free software; you can redistribute it and/or         *
# *   modify it under the terms of the GNU Library General Public           *
# *   License as published by the Free Software Foundation; either          *
# *   version 2 of the License, or (at your option) any later version.      *
# *                                                                         *
# *   This library  is distributed in the hope that it will be useful,      *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this library; see the file COPYING.LIB. If not,    *
# *   write to the Free Software Foundation, Inc., 59 Temple Place,         *
# *   Suite 330, Boston, MA  02111-1307, USA                                *
# *                                                                         *
# ***************************************************************************/

from __future__ import absolute_import

import FreeCAD as App
import Part
mm = App.Units.MilliMetre
deg = App.Units.Degree
Q = App.Units.Quantity

from AttachmentEditor.FrozenClass import FrozenClass
try:
    from Show import TempoVis
    from Show.DepGraphTools import getAllDependent
except ImportError as err:
    def TempoVis(doc):
        return None
    def getAllDependent(feature):
        return []
    App.Console.PrintWarning("AttachmentEditor: Failed to import some code from Show module. Functionality will be limited.\n")
    App.Console.PrintWarning(str(err))

if App.GuiUp:
    import FreeCADGui as Gui
    from PySide import QtCore, QtGui
    from FreeCADGui import PySideUic as uic

#-------------------------- translation-related code ----------------------------------------
#Thanks, yorik! (see forum thread "A new Part tool is being born... JoinFeatures!"
#http://forum.freecadweb.org/viewtopic.php?f=22&t=11112&start=30#p90239 )
try:
    _fromUtf8 = QtCore.QString.fromUtf8
except AttributeError:
    def _fromUtf8(s):
        return s
try:
    _encoding = QtGui.QApplication.UnicodeUTF8
    def _translate(context, text, disambig):
        return QtGui.QApplication.translate(context, text, disambig, _encoding)
except AttributeError:
    def _translate(context, text, disambig):
        return QtGui.QApplication.translate(context, text, disambig)
#--------------------------/translation-related code ----------------------------------------

def linkSubList_convertToOldStyle(references):
    ("input: [(obj1, (sub1, sub2)), (obj2, (sub1, sub2))]\n"
    "output: [(obj1, sub1), (obj1, sub2), (obj2, sub1), (obj2, sub2)]")
    result = []
    for tup in references:
        if type(tup[1]) is tuple or type(tup[1]) is list:
            for subname in tup[1]:
                result.append((tup[0], subname))
            if len(tup[1]) == 0:
                result.append((tup[0], ''))
        elif isinstance(tup[1],basestring):
            # old style references, no conversion required
            result.append(tup)
    return result


def StrFromLink(feature, subname):
    return feature.Name+ ((':'+subname) if subname else '')

def LinkFromStr(strlink, document):
    if len(strlink) == 0:
        return None
    pieces = strlink.split(':')

    feature = document.getObject(pieces[0])

    subname = ''
    if feature is None:
        raise ValueError(_translate('AttachmentEditor',"No object named {name}",None).format(name= pieces[0]))
    if len(pieces) == 2:
        subname = pieces[1]
    elif len(pieces) > 2:
        raise ValueError(_translate('AttachmentEditor',"Failed to parse link (more than one colon encountered)",None))

    return (feature,str(subname)) #wrap in str to remove unicode, which confuses assignment to PropertyLinkSubList.

def StrListFromRefs(references):
    '''input: PropertyLinkSubList. Output: list of strings for UI.'''
    references_oldstyle = linkSubList_convertToOldStyle(references)
    return [StrFromLink(feature,subelement) for (feature, subelement) in references_oldstyle]

def RefsFromStrList(strings, document):
    '''input: strings as from UI. Output: list of tuples that can be assigned to PropertyLinkSubList.'''
    refs = []
    for st in strings:
        lnk = LinkFromStr(st, document)
        if lnk is not None:
            refs.append(lnk)
    return refs

def GetSelectionAsLinkSubList():
    sel = Gui.Selection.getSelectionEx()
    result = []
    for selobj in sel:
        for subname in selobj.SubElementNames:
            result.append((selobj.Object, subname))
        if len(selobj.SubElementNames) == 0:
            result.append((selobj.Object, ''))
    return result


def PlacementsFuzzyCompare(plm1, plm2):
    pos_eq = (plm1.Base - plm2.Base).Length < 1e-7   # 1e-7 is OCC's Precision::Confusion

    q1 = plm1.Rotation.Q
    q2 = plm2.Rotation.Q
    # rotations are equal if q1 == q2 or q1 == -q2.
    # Invert one of Q's if their scalar product is negative, before comparison.
    if q1[0]*q2[0] + q1[1]*q2[1] + q1[2]*q2[2] + q1[3]*q2[3] < 0:
        q2 = [-v for v in q2]
    rot_eq = (  abs(q1[0]-q2[0]) +
                abs(q1[1]-q2[1]) +
                abs(q1[2]-q2[2]) +
                abs(q1[3]-q2[3])  ) < 1e-12   # 1e-12 is OCC's Precision::Angular (in radians)
    return pos_eq and rot_eq

class CancelError(Exception):
    def __init__(self):
        self.message = 'Canceled by user'
        self.isCancelError = True

class AttachmentEditorTaskPanel(FrozenClass):
    '''The editmode TaskPanel for attachment editing'''
    KEYmode = QtCore.Qt.ItemDataRole.UserRole # Key to use in Item.data(key) to obtain a mode associated with list item
    KEYon = QtCore.Qt.ItemDataRole.UserRole + 1 # Key to use in Item.data(key) to obtain if the mode is valid

    def __define_attributes(self):
        self.obj = None #feature being attached
        self.attacher = None #AttachEngine that is being actively used by the dialog. Its parameters are constantly and actively kept in sync with the dialog.
        self.obj_is_attachable = True # False when editing non-attachable objects (alignment, not attachment)

        self.last_sugr = None #result of last execution of suggestor

        self.form = None #Qt widget of dialog interface
        self.block = False #when True, event handlers return without doing anything (instead of doing-undoing blockSignals to everything)
        self.refLines = [] #reference lineEdit widgets, packed into a list for convenience
        self.refButtons = [] #buttons next to reference lineEdits
        self.attachmentOffsetEdits = [] #all edit boxes related to attachmentOffset
        self.i_active_ref = -1 #index of reference being selected (-1 means no reaction to selecting)
        self.auto_next = False #if true, references being selected are appended ('Selecting' state is automatically advanced to next button)

        self.tv = None #TempoVis class instance

        self.create_transaction = True # if false, dialog doesn't mess with transactions.
        self.callback_OK        = None
        self.callback_Cancel    = None
        self.callback_Apply     = None

        self._freeze()

    def __init__(self, obj_to_attach,
                       take_selection = False,
                       create_transaction = True,
                       callback_OK        = None,
                       callback_Cancel    = None,
                       callback_Apply     = None):

        self.__define_attributes()

        self.create_transaction = create_transaction
        self.callback_OK        = callback_OK
        self.callback_Cancel    = callback_Cancel
        self.callback_Apply     = callback_Apply

        self.obj = obj_to_attach
        if hasattr(obj_to_attach,'Attacher'):
            self.attacher = obj_to_attach.Attacher
        elif hasattr(obj_to_attach,'AttacherType'):
            self.attacher = Part.AttachEngine(obj_to_attach.AttacherType)
        else:
            movable = True
            if not hasattr(self.obj, "Placement"):
                movable = False
            if 'Hidden' in self.obj.getEditorMode("Placement") or 'ReadOnly' in self.obj.getEditorMode("Placement"):
                movable = False
            if not movable:
                if self.callback_Cancel:
                    self.callback_Cancel()
                raise ValueError(_translate('AttachmentEditor',"Object {name} is neither movable nor attachable, can't edit attachment",None)
                                 .format(name= self.obj.Label))

            self.obj_is_attachable = False
            self.attacher = Part.AttachEngine()

            mb = QtGui.QMessageBox()
            mb.setIcon(mb.Icon.Warning)
            mb.setText(_translate('AttachmentEditor',
                         "{obj} is not attachable. You can still use attachment editor dialog to align the object, but the attachment won't be parametric."
                         ,None)
                       .format(obj= obj_to_attach.Label))
            mb.setWindowTitle(_translate('AttachmentEditor',"Attachment",None))
            btnAbort = mb.addButton(QtGui.QMessageBox.StandardButton.Abort)
            btnOK = mb.addButton(_translate('AttachmentEditor',"Continue",None),QtGui.QMessageBox.ButtonRole.ActionRole)
            mb.setDefaultButton(btnOK)
            mb.exec_()
            if mb.clickedButton() is btnAbort:
                if self.callback_Cancel:
                    self.callback_Cancel()
                raise CancelError()

        import os
        self.form=uic.loadUi(os.path.dirname(__file__) + os.path.sep + 'TaskAttachmentEditor.ui')
        self.form.setWindowIcon(QtGui.QIcon(':/icons/Part_Attachment.svg'))
        self.form.setWindowTitle(_translate('AttachmentEditor',"Attachment",None))

        self.form.attachmentOffsetX.setProperty("unit", "mm")
        self.form.attachmentOffsetY.setProperty("unit", "mm")
        self.form.attachmentOffsetZ.setProperty("unit", "mm")
        Gui.ExpressionBinding(self.form.attachmentOffsetX).bind(self.obj,"AttachmentOffset.Base.x")
        Gui.ExpressionBinding(self.form.attachmentOffsetY).bind(self.obj,"AttachmentOffset.Base.y")
        Gui.ExpressionBinding(self.form.attachmentOffsetZ).bind(self.obj,"AttachmentOffset.Base.z")

        Gui.ExpressionBinding(self.form.attachmentOffsetYaw).bind(self.obj,"AttachmentOffset.Rotation.Yaw")
        Gui.ExpressionBinding(self.form.attachmentOffsetPitch).bind(self.obj,"AttachmentOffset.Rotation.Pitch")
        Gui.ExpressionBinding(self.form.attachmentOffsetRoll).bind(self.obj,"AttachmentOffset.Rotation.Roll")

        self.refLines = [self.form.lineRef1,
                         self.form.lineRef2,
                         self.form.lineRef3,
                         self.form.lineRef4]
        self.refButtons = [self.form.buttonRef1,
                           self.form.buttonRef2,
                           self.form.buttonRef3,
                           self.form.buttonRef4]
        self.attachmentOffsetEdits = [self.form.attachmentOffsetX,
                                    self.form.attachmentOffsetY,
                                    self.form.attachmentOffsetZ,
                                    self.form.attachmentOffsetYaw,
                                    self.form.attachmentOffsetPitch,
                                    self.form.attachmentOffsetRoll]

        self.block = False

        for i in range(len(self.refLines)):
            QtCore.QObject.connect(self.refLines[i], QtCore.SIGNAL('textEdited(QString)'), lambda txt, i=i: self.lineRefChanged(i,txt))

        for i in range(len(self.refLines)):
            QtCore.QObject.connect(self.refButtons[i], QtCore.SIGNAL('clicked()'), lambda i=i: self.refButtonClicked(i))

        for i in range(len(self.attachmentOffsetEdits)):
            QtCore.QObject.connect(self.attachmentOffsetEdits[i], QtCore.SIGNAL('valueChanged(double)'), lambda val, i=i: self.attachmentOffsetChanged(i,val))

        QtCore.QObject.connect(self.form.checkBoxFlip, QtCore.SIGNAL('clicked()'), self.checkBoxFlipClicked)

        QtCore.QObject.connect(self.form.listOfModes, QtCore.SIGNAL('itemSelectionChanged()'), self.modeSelected)

        if self.create_transaction:
            self.obj.Document.openTransaction(_translate('AttachmentEditor',"Edit attachment of {feat}",None).format(feat= self.obj.Name))


        self.readParameters()


        if len(self.attacher.References) == 0 and take_selection:
            sel = GetSelectionAsLinkSubList()
            for i in range(len(sel))[::-1]:
                if sel[i][0] is obj_to_attach:
                    sel.pop(i)
            self.attacher.References = sel
            # need to update textboxes
            self.fillAllRefLines()

        if len(self.attacher.References) == 0:
            self.i_active_ref = 0
            self.auto_next = True
        else:
            self.i_active_ref = -1
            self.auto_next = False

        Gui.Selection.addObserver(self)

        self.updatePreview()
        self.updateRefButtons()

        self.tv = TempoVis(self.obj.Document, tag= "PartGui.TaskAttachmentEditor")
        if self.tv: # tv will still be None if Show module is unavailable
            self.tv.hide_all_dependent(self.obj)
            self.tv.show(self.obj)
            self.tv.setUnpickable(self.obj)
            self.tv.modifyVPProperty(self.obj, "Transparency", 70)
            self.tv.show([obj for (obj,subname) in self.attacher.References])

    # task dialog handling
    def getStandardButtons(self):
        return int(QtGui.QDialogButtonBox.Ok) | int(QtGui.QDialogButtonBox.Cancel)| int(QtGui.QDialogButtonBox.Apply)

    def clicked(self,button):
        if button == QtGui.QDialogButtonBox.Apply:
            if self.obj_is_attachable:
                self.writeParameters()
            if self.create_transaction:
                self.obj.Document.commitTransaction()
                self.obj.Document.openTransaction(_translate('AttachmentEditor',"Edit attachment of {feat}",None).format(feat= self.obj.Name))
            self.updatePreview()
            if self.callback_Apply:
                self.callback_Apply()

    def accept(self):
        if self.obj_is_attachable:
            self.writeParameters()
        if self.create_transaction:
            self.obj.Document.commitTransaction()
        self.cleanUp()
        Gui.Control.closeDialog()
        if self.callback_OK:
            self.callback_OK()

    def reject(self):
        if self.create_transaction:
            self.obj.Document.abortTransaction()
        self.cleanUp()
        Gui.Control.closeDialog()
        if self.callback_Cancel:
            self.callback_Cancel()


    #selectionObserver stuff
    def addSelection(self,docname,objname,subname,pnt):
        i = self.i_active_ref
        if i < 0:
            #not selecting any reference
            return
        if i > 0 and self.auto_next:
            prevref = LinkFromStr( self.refLines[i-1].text(), self.obj.Document )
            if prevref[0].Name == objname and subname == '':
                # whole object was selected by double-clicking
                # its subelement was already written to line[i-1], so we decrease i to overwrite the lineRefChanged
                i -= 1
        if i > len(self.refLines)-1:
            # all 4 references have been selected, finish
            assert(self.auto_next)
            self.i_active_ref = -1
            self.updateRefButtons()
            return
        if i > -1:
            # assign the selected reference
            if objname == self.obj.Name:
                self.form.message.setText(_translate('AttachmentEditor',"Ignored. Can't attach object to itself!",None))
                return
            if App.getDocument(docname).getObject(objname) in getAllDependent(self.obj):
                self.form.message.setText(_translate('AttachmentEditor',"{obj1} depends on object being attached, can't use it for attachment",None).format(obj1= objname))
                return

            self.refLines[i].setText( StrFromLink(App.getDocument(docname).getObject(objname), subname) )
            self.lineRefChanged(i,'')
            if self.auto_next:
                i += 1
        self.i_active_ref = i
        self.updateRefButtons()

    # slots

    def attachmentOffsetChanged(self, index, value):
        if self.block:
            return
        plm = self.attacher.AttachmentOffset
        pos = plm.Base
        if index==0:
            pos.x = Q(self.form.attachmentOffsetX.text()).getValueAs(mm)
        if index==1:
            pos.y = Q(self.form.attachmentOffsetY.text()).getValueAs(mm)
        if index==2:
            pos.z = Q(self.form.attachmentOffsetZ.text()).getValueAs(mm)
        if index >= 0  and  index <= 2:
            plm.Base = pos

        rot = plm.Rotation;
        (yaw, pitch, roll) = rot.toEuler()
        if index==3:
            yaw = Q(self.form.attachmentOffsetYaw.text()).getValueAs(deg)
        if index==4:
            pitch = Q(self.form.attachmentOffsetPitch.text()).getValueAs(deg)
        if index==5:
            roll = Q(self.form.attachmentOffsetRoll.text()).getValueAs(deg)
        if index >= 3  and  index <= 5:
            rot = App.Rotation(yaw,pitch,roll)
            plm.Rotation = rot

        self.attacher.AttachmentOffset = plm

        self.updatePreview()

    def checkBoxFlipClicked(self):
        if self.block:
            return
        self.attacher.Reverse = self.form.checkBoxFlip.isChecked()
        self.updatePreview()

    def lineRefChanged(self, index, value):
        if self.block:
            return
        # not parsing links here, because doing it in updatePreview will display error message
        self.updatePreview()

    def refButtonClicked(self, index):
        if self.block:
            return
        if self.i_active_ref == index:
            #stop selecting
            self.i_active_ref = -1
        else:
            #start selecting
            self.i_active_ref = index
            self.auto_next = False
        self.updateRefButtons()

    def modeSelected(self):
        if self.block:
            return
        self.attacher.Mode = self.getCurrentMode()
        self.updatePreview()

    #internal methods
    def writeParameters(self):
        'Transfer from the dialog to the object'
        self.attacher.writeParametersToFeature(self.obj)

    def readParameters(self):
        'Transfer from the object to the dialog'
        if self.obj_is_attachable:
            self.attacher.readParametersFromFeature(self.obj)

        plm = self.attacher.AttachmentOffset
        try:
            old_selfblock = self.block
            self.block = True
            self.form.attachmentOffsetX.lineEdit().setText    ((plm.Base.x * mm).UserString)
            self.form.attachmentOffsetY.lineEdit().setText    ((plm.Base.y * mm).UserString)
            self.form.attachmentOffsetZ.lineEdit().setText    ((plm.Base.z * mm).UserString)
            self.form.attachmentOffsetYaw.lineEdit().setText  ((plm.Rotation.toEuler()[0] * deg).UserString)
            self.form.attachmentOffsetPitch.lineEdit().setText((plm.Rotation.toEuler()[1] * deg).UserString)
            self.form.attachmentOffsetRoll.lineEdit().setText ((plm.Rotation.toEuler()[2] * deg).UserString)

            self.form.checkBoxFlip.setChecked(self.attacher.Reverse)

            self.fillAllRefLines()
        finally:
            self.block = old_selfblock

    def fillAllRefLines(self):
        old_block = self.block
        try:
            self.block = True
            strings = StrListFromRefs(self.attacher.References)
            if len(strings) < len(self.refLines):
                strings.extend(['']*(len(self.refLines) - len(strings)))
            for i in range(len(self.refLines)):
                self.refLines[i].setText(strings[i])
        finally:
                self.block = old_block

    def parseAllRefLines(self):
        self.attacher.References = RefsFromStrList([le.text() for le in self.refLines], self.obj.Document)

    def updateListOfModes(self):
        '''needs suggestor to have been called, and assigned to self.last_sugr'''
        try:
            old_selfblock = self.block
            self.block = True
            list_widget = self.form.listOfModes
            list_widget.clear()
            sugr = self.last_sugr
            # always have the option to choose Deactivated mode
            valid_modes = ['Deactivated'] + sugr['allApplicableModes']

            # add valid modes
            for m in valid_modes:
                item = QtGui.QListWidgetItem()
                txt = self.attacher.getModeInfo(m)['UserFriendlyName']
                item.setText(txt)
                item.setData(self.KEYmode,m)
                item.setData(self.KEYon,True)
                if m == sugr['bestFitMode']:
                    f = item.font()
                    f.setBold(True)
                    item.setFont(f)
                list_widget.addItem(item)
                item.setSelected(self.attacher.Mode == m)
            # add potential modes
            for m in sugr['reachableModes'].keys():
                item = QtGui.QListWidgetItem()
                txt = self.attacher.getModeInfo(m)['UserFriendlyName']
                listlistrefs = sugr['reachableModes'][m]
                if len(listlistrefs) == 1:
                    listrefs_userfriendly = [self.attacher.getRefTypeInfo(t)['UserFriendlyName'] for t in listlistrefs[0]]
                    txt = _translate('AttachmentEditor',"{mode} (add {morerefs})",None).format(mode= txt,
                                                           morerefs= u"+".join(listrefs_userfriendly))
                else:
                    txt = _translate('AttachmentEditor',"{mode} (add more references)",None).format(mode= txt)
                item.setText(txt)
                item.setData(self.KEYmode,m)
                item.setData(self.KEYon,True)
                if m == sugr['bestFitMode']:
                    f = item.font()
                    f.setBold(True)
                    item.setFont(f)

                #disable this item
                f = item.flags()
                f = f & ~(QtCore.Qt.ItemFlag.ItemIsEnabled | QtCore.Qt.ItemFlag.ItemIsSelectable)
                item.setFlags(f)

                list_widget.addItem(item)

            # re-scan the list to fill in tooltips
            for item in list_widget.findItems('', QtCore.Qt.MatchContains):
                m = item.data(self.KEYmode)
                on = item.data(self.KEYon)

                mi = self.attacher.getModeInfo(m)
                cmb = []
                for refstr in mi['ReferenceCombinations']:
                    refstr_userfriendly = [self.attacher.getRefTypeInfo(t)['UserFriendlyName'] for t in refstr]
                    cmb.append(u", ".join(refstr_userfriendly))

                tip = mi['BriefDocu']
                if (m != 'Deactivated'):
                    tip += u"\n\n"
                    tip += _translate('AttachmentEditor', "Reference combinations:", None) + u" \n\n".join(cmb)

                item.setToolTip(tip)

        finally:
            self.block = old_selfblock


    def updateRefButtons(self):
        try:
            old_selfblock = self.block
            self.block = True
            for i in range(len(self.refButtons)):
                btn = self.refButtons[i]
                btn.setCheckable(True)
                btn.setChecked(self.i_active_ref == i)
                typ = _translate('AttachmentEditor',"Reference{i}",None).format(i= str(i+1))
                if self.last_sugr is not None:
                    typestr = self.last_sugr['references_Types']
                    if i < len(typestr):
                        typ = self.attacher.getRefTypeInfo(typestr[i])['UserFriendlyName']
                btn.setText(_translate('AttachmentEditor',"Selecting...",None) if self.i_active_ref == i else typ)
        finally:
            self.block = old_selfblock

    def getCurrentMode(self):
        list_widget = self.form.listOfModes
        sel = list_widget.selectedItems()
        if len(sel) == 1:
            if sel[0].data(self.KEYon):
                return str(sel[0].data(self.KEYmode)) # data() returns unicode, which confuses attacher
        # nothing selected in list. Return suggested
        if self.last_sugr is not None:
            if self.last_sugr['message'] == 'OK':
                return self.last_sugr['bestFitMode']
        # no suggested mode. Return current, so it doesn't change
        return self.attacher.Mode

    def updatePreview(self):
        new_plm = None

        try:
            self.parseAllRefLines()
            self.last_sugr = self.attacher.suggestModes()
            if self.last_sugr['message'] == 'LinkBroken':
                raise ValueError(_translate('AttachmentEditor',"Failed to resolve links. {err}",None).format(err= self.last_sugr['error']))

            self.updateListOfModes()

            self.attacher.Mode = self.getCurrentMode()

            new_plm = self.attacher.calculateAttachedPlacement(self.obj.Placement)
            if new_plm is None:
                self.form.message.setText(_translate('AttachmentEditor',"Not attached",None))
            else:
                self.form.message.setText(    _translate('AttachmentEditor',"Attached with mode {mode}",None)
                                              .format(  mode=   self.attacher.getModeInfo(self.getCurrentMode())['UserFriendlyName']  )    )
                if PlacementsFuzzyCompare(self.obj.Placement, new_plm) == False:
                    # assign only if placement changed. this avoids touching the object
                    # when entering and extiting dialog without changing anything
                    self.obj.Placement = new_plm
        except Exception as err:
            self.form.message.setText(_translate('AttachmentEditor',"Error: {err}",None).format(err= str(err)))

        if new_plm is not None:
            self.form.groupBox_AttachmentOffset.setTitle(_translate('AttachmentEditor',"Attachment Offset (in local coordinates):",None))
            self.form.groupBox_AttachmentOffset.setEnabled(True)
        else:
            self.form.groupBox_AttachmentOffset.setTitle(_translate('AttachmentEditor',"Attachment Offset (inactive - not attached):",None))
            self.form.groupBox_AttachmentOffset.setEnabled(False)

    def cleanUp(self):
        '''stuff that needs to be done when dialog is closed.'''
        Gui.Selection.removeObserver(self)
        if self.tv:
            self.tv.restore()



# -*- coding: utf-8 -*-

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2017 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import PathScripts.PathGeom as PathGeom
import PathScripts.PathLog as PathLog
import PathScripts.PathUtil as PathUtil
import PySide


__title__ = "Path UI helper and utility functions"
__author__ = "sliptonic (Brad Collette)"
__url__ = "http://www.freecadweb.org"
__doc__ = "A collection of helper and utility functions for the Path GUI."

def translate(context, text, disambig=None):
    return PySide.QtCore.QCoreApplication.translate(context, text, disambig)

LOGLEVEL = False

if LOGLEVEL:
    PathLog.setLevel(PathLog.Level.DEBUG, PathLog.thisModule())
    PathLog.trackModule(PathLog.thisModule())
else:
    PathLog.setLevel(PathLog.Level.INFO, PathLog.thisModule())


def updateInputField(obj, prop, widget, onBeforeChange=None):
    '''updateInputField(obj, prop, widget) ... update obj's property prop with the value of widget.
    The property's value is only assigned if the new value differs from the current value.
    This prevents onChanged notifications where the value didn't actually change.
    Gui::InputField and Gui::QuantitySpinBox widgets are supported - and the property can
    be of type Quantity or Float.
    If onBeforeChange is specified it is called before a new value is assigned to the property.
    Returns True if a new value was assigned, False otherwise (new value is the same as the current).
    '''
    value = FreeCAD.Units.Quantity(widget.text()).Value
    attr = PathUtil.getProperty(obj, prop)
    attrValue = attr.Value if hasattr(attr, 'Value') else attr

    isDiff = False
    if not PathGeom.isRoughly(attrValue, value):
        isDiff = True
    else:
        if hasattr(obj, 'ExpressionEngine'):
            noExpr = True
            for (prp, expr) in obj.ExpressionEngine:
                if prp == prop:
                    noExpr = False
                    PathLog.debug('prop = "expression": {} = "{}"'.format(prp, expr))
                    value = FreeCAD.Units.Quantity(obj.evalExpression(expr)).Value
                    if not PathGeom.isRoughly(attrValue, value):
                        isDiff = True
                    break
            if noExpr:
                widget.setReadOnly(False)
                widget.setStyleSheet("color: black")
            else:
                widget.setReadOnly(True)
                widget.setStyleSheet("color: gray")
            widget.update()

    if isDiff:
        PathLog.debug("updateInputField(%s, %s): %.2f -> %.2f" % (obj.Label, prop, attr, value))
        if onBeforeChange:
            onBeforeChange(obj)
        PathUtil.setProperty(obj, prop, value)
        return True

    return False


class QuantitySpinBox:
    '''Controller class to interface a Gui::QuantitySpinBox.
    The spin box gets bound to a given property and supports update in both directions.
    QuatitySpinBox(widget, obj, prop, onBeforeChange=None)
            widget ... expected to be reference to a Gui::QuantitySpinBox
            obj    ... document object
            prop   ... canonical name of the (sub-) property
            onBeforeChange ... an optional callback being executed before the value of the property is changed
    '''

    def __init__(self, widget, obj, prop, onBeforeChange=None):
        self.obj = obj
        self.widget = widget
        self.prop = prop
        self.onBeforeChange = onBeforeChange

        attr = PathUtil.getProperty(self.obj, self.prop)
        if attr is not None:
            if hasattr(attr, 'Value'):
                widget.setProperty('unit', attr.getUserPreferred()[2])
            widget.setProperty('binding', "%s.%s" % (obj.Name, prop))
            self.valid = True
        else:
            PathLog.warning(translate('PathGui', "Cannot find property %s of %s") % (prop, obj.Label))
            self.valid = False

    def expression(self):
        '''expression() ... returns the expression if one is bound to the property'''
        if self.valid:
            return self.widget.property('expression')
        return ''

    def setMinimum(self, quantity):
        if self.valid:
            value = quantity.Value if hasattr(quantity, 'Value') else quantity
            self.widget.setProperty('setMinimum', value)

    def updateSpinBox(self, quantity=None):
        '''updateSpinBox(quantity=None) ... update the display value of the spin box.
        If no value is provided the value of the bound property is used.
        quantity can be of type Quantity or Float.'''
        if self.valid:
            if quantity is None:
                quantity = PathUtil.getProperty(self.obj, self.prop)
            value = quantity.Value if hasattr(quantity, 'Value') else quantity
            self.widget.setProperty('rawValue', value)

    def updateProperty(self):
        '''updateProperty() ... update the bound property with the value from the spin box'''
        if self.valid:
            return updateInputField(self.obj, self.prop, self.widget, self.onBeforeChange)
        return None



#/***************************************************************************
# *   Copyright (c) 2016 Victor Titov (DeepSOIC) <vv.titov@gmail.com>       *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This library is free software; you can redistribute it and/or         *
# *   modify it under the terms of the GNU Library General Public           *
# *   License as published by the Free Software Foundation; either          *
# *   version 2 of the License, or (at your option) any later version.      *
# *                                                                         *
# *   This library  is distributed in the hope that it will be useful,      *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this library; see the file COPYING.LIB. If not,    *
# *   write to the Free Software Foundation, Inc., 59 Temple Place,         *
# *   Suite 330, Boston, MA  02111-1307, USA                                *
# *                                                                         *
# ***************************************************************************/

__title__ = "BOPTools package"
__url__ = "http://www.freecadweb.org"
__doc__ = """BOPTools Package (part of FreeCAD). Routines that power Connect, Embed, Cutout,
BooleanFragments, Slice and XOR features of Part Workbench. Useful for other custom
BOP-like operations"""

## @package BOPTools
#  \ingroup PART

__all__ = [
"GeneralFuseResult",
"JoinAPI",
"JoinFeatures",
"ShapeMerge",
"Utils",
"SplitAPI",
"SplitFeatures",
]

def importAll():
    "importAll(): imports all modules of BOPTools package"
    from . import GeneralFuseResult
    from . import JoinAPI
    from . import JoinFeatures
    from . import ShapeMerge
    from . import Utils
    from . import SplitAPI
    from . import SplitFeatures

def reloadAll():
    "reloadAll(): reloads all modules of BOPTools package. Useful for debugging."
    for modstr in __all__:
        reload(globals()[modstr])
    import FreeCAD
    if FreeCAD.GuiUp:
        addCommands()

def addCommands():
    "addCommands(): add all GUI commands of BOPTools package to FreeCAD command manager."
    JoinFeatures.addCommands()
    SplitFeatures.addCommands()

def generalFuseIsAvailable():
    """generalFuseIsAvailable(): returns True if FreeCAD's Part.Shape.generalFuse is functional.
    True if Part.OCC_VERSION >= 6.9.0."""
    import Part
    if not hasattr(Part, "OCC_VERSION"):
        return False
    else:
        ver_string = Part.OCC_VERSION
        import re
        match = re.match(r"([0-9]+)\.([0-9]+)\.([0-9]+)",ver_string)
        major,minor,rev = match.groups()
        major = int(major); minor = int(minor); rev = int(rev)
        return (major,minor,rev)>=(6,9,0)



# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2020 Carlo Pavan <carlopav@gmail.com>                   *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Draft Statusbar commands.

This module provide the code for the Draft Statusbar, activated by initGui
"""
## @package init_draft_statusbar
# \ingroup DRAFT
# \brief This module provides the code for the Draft Statusbar.

from PySide import QtCore
from PySide import QtGui
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import FreeCADGui as Gui

from draftutils.init_tools import get_draft_snap_commands

#----------------------------------------------------------------------------
# SCALE WIDGET FUNCTIONS
#----------------------------------------------------------------------------

draft_scales_metrics =  ["1:1000", "1:500", "1:250", "1:200", "1:100", 
                         "1:50", "1:25","1:20", "1:10", "1:5","1:2",
                         "1:1",
                         "2:1", "5:1", "10:1", "20:1",                 
                         QT_TRANSLATE_NOOP("draft","custom"),                 
                        ]

draft_scales_arch_imperial =  ["1/16in=1ft", "3/32in=1ft", "1/8in=1ft",
                               "3/16in=1ft", "1/4in=1ft","3/8in=1ft", 
                               "1/2in=1ft", "3/4in=1ft", "1in=1ft", 
                               "1.5in=1ft", "3in=1ft",    
                               QT_TRANSLATE_NOOP("draft","custom"),
                              ]

draft_scales_eng_imperial =  ["1in=10ft", "1in=20ft", "1in=30ft",
                              "1in=40ft", "1in=50ft", "1in=60ft", 
                              "1in=70ft", "1in=80ft", "1in=90ft", 
                              "1in=100ft",    
                              QT_TRANSLATE_NOOP("draft","custom"),
                             ]

def get_scales(unit_system = 0):
    """
    returns the list of preset scales accordin to unit system.
    
    Parameters:
    unit_system =   0 : default from user preferences
                    1 : metrics
                    2 : imperial architectural
                    3 : imperial engineering
    """

    if unit_system == 0:
        param = App.ParamGet("User parameter:BaseApp/Preferences/Units")
        scale_units_system = param.GetInt("UserSchema", 0)
        if scale_units_system in [0, 1, 4, 6]:
            return draft_scales_metrics
        elif scale_units_system in [2, 3, 5]:
            return draft_scales_arch_imperial
        elif scale_units_system in [7]:
            return draft_scales_eng_imperial
    elif unit_system == 1:
        return draft_scales_metrics
    elif unit_system == 2:
        return draft_scales_arch_imperial
    elif unit_system == 3:
        return draft_scales_eng_imperial


def scale_to_label(scale):
    """
    transform a float number into a 1:X or X:1 scale and return it as label
    """
    f = 1/scale
    f = round(f,2)
    f = f.as_integer_ratio()
    if f[1] == 1 or f[0] == 1:
        label = str(f[1]) + ":" + str(f[0])
        return label
    else:
        return str(scale)
    
def label_to_scale(label):
    """
    transform a scale string into scale factor as float
    """
    try :
        scale = float(label)
        return scale
    except :
        if ":" in label:
            f = label.split(":")
        elif "=" in label:
            f = label.split("=")
        else:
            return
        if len(f) == 2:
            try:
                num = App.Units.Quantity(f[0]).Value
                den = App.Units.Quantity(f[1]).Value
                scale = num/den
                return scale
            except:
                err = QT_TRANSLATE_NOOP("draft",
                                        "Unable to convert input into a "
                                        "scale factor")
                App.Console.PrintWarning(err)
                return None

def _set_scale(action):
    """
    triggered by scale pushbutton, set DraftAnnotationScale in preferences
    """
    # set the label of the scale button
    param = App.ParamGet("User parameter:BaseApp/Preferences/Mod/Draft")
    
    mw = Gui.getMainWindow()
    sb = mw.statusBar()
    scale_widget = sb.findChild(QtGui.QToolBar,"draft_status_scale_widget")
    if action.text() == QT_TRANSLATE_NOOP("draft","custom"):
        dialog_text = QT_TRANSLATE_NOOP("draft",
                                        "Set custom annotation scale in "
                                        "format x:x, x=x"
                                       )
        custom_scale = QtGui.QInputDialog.getText(None, "Set custom scale",
                                                  dialog_text)
        if custom_scale[1]:
            print(custom_scale[0])
            scale = label_to_scale(custom_scale[0])
            if scale is None:
                return
            param.SetFloat("DraftAnnotationScale", scale)
            cs = scale_to_label(scale)
            scale_widget.scaleLabel.setText(cs)
    else:
        text_scale = action.text()
        scale_widget.scaleLabel.setText(text_scale)
        scale = label_to_scale(text_scale)
        param.SetFloat("DraftAnnotationScale", scale)

#----------------------------------------------------------------------------
# MAIN DRAFT STATUSBAR FUNCTIONS
#----------------------------------------------------------------------------

def init_draft_statusbar(sb):
    """
    this function initializes draft statusbar
    """
    param = App.ParamGet("User parameter:BaseApp/Preferences/Mod/Draft")

    # SNAP WIDGET - init ----------------------------------------------------
    
    snap_widget = QtGui.QToolBar()
    snap_widget.setObjectName("draft_snap_widget")
    snap_widget.setIconSize(QtCore.QSize(16,16))
    
    # GRID BUTTON - init
    gridbutton = QtGui.QPushButton(snap_widget)
    gridbutton.setIcon(QtGui.QIcon.fromTheme("Draft",
                                             QtGui.QIcon(":/icons/"
                                                         "Draft_Grid.svg")))
    gridbutton.setToolTip(QT_TRANSLATE_NOOP("Draft",
                                            "Toggles Grid On/Off"))
    gridbutton.setObjectName("Grid_Statusbutton")
    gridbutton.setWhatsThis("Draft_ToggleGrid")
    gridbutton.setFlat(True)
    QtCore.QObject.connect(gridbutton,QtCore.SIGNAL("clicked()"),
                           lambda f=Gui.doCommand, 
                           arg='Gui.runCommand("Draft_ToggleGrid")':f(arg))
    snap_widget.addWidget(gridbutton)

    # SNAP BUTTON - init
    snappref = param.GetString("snapModes","111111111101111")[0]
    snapbutton = QtGui.QPushButton(snap_widget)
    snapbutton.setIcon(QtGui.QIcon.fromTheme("Draft",
                                              QtGui.QIcon(":/icons/"
                                                          "Snap_Lock.svg")))
    snapbutton.setObjectName("Snap_Statusbutton")
    snapbutton.setWhatsThis("Draft_ToggleLockSnap")
    snapbutton.setToolTip(QT_TRANSLATE_NOOP("Draft",
                                          "Object snapping"))
    snapbutton.setCheckable(True)
    snapbutton.setChecked(bool(int(snappref)))
    snapbutton.setFlat(True)

    snaps_menu = QtGui.QMenu(snapbutton)
    snaps_menu.setObjectName("draft_statusbar_snap_toolbar")
    
    snap_gui_commands = get_draft_snap_commands()
    if 'Draft_Snap_Ortho' in snap_gui_commands:
        snap_gui_commands.remove('Draft_Snap_Ortho')
    if 'Draft_Snap_WorkingPlane' in snap_gui_commands:
        snap_gui_commands.remove('Draft_Snap_WorkingPlane')
    if 'Draft_Snap_Dimensions' in snap_gui_commands:
        snap_gui_commands.remove('Draft_Snap_Dimensions')
    if 'Draft_ToggleGrid' in snap_gui_commands:
        snap_gui_commands.remove('Draft_ToggleGrid')

    Gui.Snapper.init_draft_snap_buttons(snap_gui_commands,snaps_menu, "_Statusbutton")
    Gui.Snapper.restore_snap_buttons_state(snaps_menu, "_Statusbutton")

    snapbutton.setMenu(snaps_menu)
    snap_widget.addWidget(snapbutton)


    # DIMENSION BUTTON - init
    dimpref = param.GetString("snapModes","111111111101111")[13]
    dimbutton = QtGui.QPushButton(snap_widget)
    dimbutton.setIcon(QtGui.QIcon.fromTheme("Draft",
                                            QtGui.QIcon(":/icons/"
                                                        "Snap_Dimensions.svg")))
    dimbutton.setToolTip(QT_TRANSLATE_NOOP("Draft",
                                           "Toggles Visual Aid Dimensions On/Off"))
    dimbutton.setObjectName("Draft_Snap_Dimensions_Statusbutton")
    dimbutton.setWhatsThis("Draft_ToggleDimensions")
    dimbutton.setFlat(True)
    dimbutton.setCheckable(True)
    dimbutton.setChecked(bool(int(dimpref)))
    QtCore.QObject.connect(dimbutton,QtCore.SIGNAL("clicked()"),
                           lambda f=Gui.doCommand, 
                           arg='Gui.runCommand("Draft_Snap_Dimensions")':f(arg))
    snap_widget.addWidget(dimbutton)

    # ORTHO BUTTON - init
    ortopref = param.GetString("snapModes","111111111101111")[10]
    orthobutton = QtGui.QPushButton(snap_widget)
    orthobutton.setIcon(QtGui.QIcon.fromTheme("Draft",
                                              QtGui.QIcon(":/icons/"
                                                          "Snap_Ortho.svg")))
    orthobutton.setObjectName("Draft_Snap_Ortho"+"_Statusbutton")
    orthobutton.setWhatsThis("Draft_ToggleOrtho")
    orthobutton.setToolTip(QT_TRANSLATE_NOOP("Draft",
                                             "Toggles Ortho On/Off"))
    orthobutton.setFlat(True)
    orthobutton.setCheckable(True)
    orthobutton.setChecked(bool(int(ortopref)))
    QtCore.QObject.connect(orthobutton,QtCore.SIGNAL("clicked()"),
                           lambda f=Gui.doCommand, 
                           arg='Gui.runCommand("Draft_Snap_Ortho")':f(arg))
    snap_widget.addWidget(orthobutton)

    # WORKINGPLANE BUTTON - init
    wppref = param.GetString("snapModes","111111111101111")[14]
    wpbutton = QtGui.QPushButton(snap_widget)
    wpbutton.setIcon(QtGui.QIcon.fromTheme("Draft",
                                              QtGui.QIcon(":/icons/"
                                                          "Snap_WorkingPlane.svg")))
    wpbutton.setObjectName("Draft_Snap_WorkingPlane_Statusbutton")
    wpbutton.setWhatsThis("Draft_ToggleWorkingPlaneSnap")
    wpbutton.setToolTip(QT_TRANSLATE_NOOP("Draft",
                                          "Toggles Constrain to Working Plane On/Off"))
    wpbutton.setFlat(True)
    wpbutton.setCheckable(True)
    wpbutton.setChecked(bool(int(wppref)))
    QtCore.QObject.connect(wpbutton,QtCore.SIGNAL("clicked()"),
                           lambda f=Gui.doCommand, 
                           arg='Gui.runCommand("Draft_Snap_WorkingPlane")':f(arg))
    snap_widget.addWidget(wpbutton)

    # add snap widget to the statusbar
    sb.insertPermanentWidget(2, snap_widget)
    snap_widget.show()

   
    # SCALE WIDGET ----------------------------------------------------------
    scale_widget = QtGui.QToolBar()
    scale_widget.setObjectName("draft_status_scale_widget")
    
    # get scales list according to system units
    draft_scales = get_scales()

    # get draft annotation scale
    draft_annotation_scale = param.GetFloat("DraftAnnotationScale", 1.0)

    # initializes scale widget
    scale_widget.draft_scales = draft_scales
    scaleLabel = QtGui.QPushButton("Scale")
    scaleLabel.setObjectName("ScaleLabel")
    scaleLabel.setFlat(True)
    menu = QtGui.QMenu(scaleLabel)
    gUnits = QtGui.QActionGroup(menu)
    for u in draft_scales:
        a = QtGui.QAction(gUnits)
        a.setText(u)
        menu.addAction(a)
    scaleLabel.setMenu(menu)
    gUnits.triggered.connect(_set_scale)
    scale_label = scale_to_label(draft_annotation_scale)
    scaleLabel.setText(scale_label)
    tooltip = "Set the scale used by draft annotation tools"
    scaleLabel.setToolTip(QT_TRANSLATE_NOOP("draft",tooltip))
    scale_widget.addWidget(scaleLabel)
    scale_widget.scaleLabel = scaleLabel
    
    # add scale widget to the statusbar
    sb.insertPermanentWidget(3, scale_widget)
    scale_widget.show()

def show_draft_statusbar():
    """
    shows draft statusbar if present or initializes it
    """
    params = App.ParamGet("User parameter:BaseApp/Preferences/Mod/Draft")
    display_statusbar = params.GetBool("DisplayStatusbar", True)

    if not display_statusbar:
        return

    mw = Gui.getMainWindow()
    if mw:
        sb = mw.statusBar()

        scale_widget = sb.findChild(QtGui.QToolBar,
                                    "draft_status_scale_widget")
        if scale_widget:
            scale_widget.show()
        elif params.GetBool("DisplayStatusbarScaleWidget", True):
            init_draft_statusbar(sb)
            
        snap_widget = sb.findChild(QtGui.QToolBar,"draft_snap_widget")
        if snap_widget:
            snap_widget.show()
        elif params.GetBool("DisplayStatusbarSnapWidget", True):
            init_draft_statusbar(sb)


def hide_draft_statusbar():
    """
    hides draft statusbar if present
    """
    mw = Gui.getMainWindow()
    if not mw:
        return
    sb = mw.statusBar()
    
    # hide scale widget
    scale_widget = sb.findChild(QtGui.QToolBar,
                                "draft_status_scale_widget")
    if scale_widget:
        scale_widget.hide()
    else:
        # when switching workbenches, the toolbar sometimes "jumps"
        # out of the status bar to any other dock area...
        scale_widget = mw.findChild(QtGui.QToolBar,
                                    "draft_status_scale_widget")
        if scale_widget:
            scale_widget.hide()
            
    # hide snap widget
    snap_widget = sb.findChild(QtGui.QToolBar,"draft_snap_widget")
    if snap_widget:
        snap_widget.hide()
    else:
        # when switching workbenches, the toolbar sometimes "jumps"
        # out of the status bar to any other dock area...
        snap_widget = mw.findChild(QtGui.QToolBar,"draft_snap_widget")
        if snap_widget:
            snap_widget.hide()                


# -*- coding: utf-8 -*-
#/******************************************************************************
# *   Copyright (c) 2012 Jan Rheinländer <jrheinlaender@users.sourceforge.net> *
# *                                                                            *
# *   This file is part of the FreeCAD CAx development system.                 *
# *                                                                            *
# *   This library is free software; you can redistribute it and/or            *
# *   modify it under the terms of the GNU Library General Public              *
# *   License as published by the Free Software Foundation; either             *
# *   version 2 of the License, or (at your option) any later version.         *
# *                                                                            *
# *   This library  is distributed in the hope that it will be useful,         *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of           *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *
# *   GNU Library General Public License for more details.                     *
# *                                                                            *
# *   You should have received a copy of the GNU Library General Public        *
# *   License along with this library; see the file COPYING.LIB. If not,       *
# *   write to the Free Software Foundation, Inc., 59 Temple Place,            *
# *   Suite 330, Boston, MA  02111-1307, USA                                   *
# *                                                                            *
# ******************************************************************************/

import FreeCAD, FreeCADGui
from .SegmentFunction import SegmentFunction,  IntervalFunction,  StressFunction,  TranslationFunction
from .ShaftFeature import ShaftFeature
from .ShaftDiagram import Diagram
import math

class ShaftSegment:
    def __init__(self, l, d, di):
        self.length = l
        self.diameter = d
        self.innerdiameter = di
        self.constraintType = "None"
        self.constraint = None

class Shaft:
    "The axis of the shaft is always assumed to correspond to the X-axis"
    # Names (note Qy corresponds with Mz, and Qz with My)
    Fstr = ["Nx","Qy","Qz"] # Forces
    Mstr = ["Mx","Mz","My"] # Moments
    wstr = ["",  "wy",  "wz"] # Translations
    sigmaNstr = ["sigmax","sigmay","sigmaz"] # Normal/shear stresses
    sigmaBstr = ["taut","sigmabz", "sigmaby"] # Torsion/bending stresses
    # For diagram labeling
    Qstrings = (("Normal force [x]",  "x", "mm", "N_x",  "N"), 
                        ("Shear force [y]",  "x",  "mm",  "Q_y",  "N"), 
                        ("Shear force [z]",  "x",  "mm",  "Q_z",  "N"))
    Mstrings = (("Torque [x]",  "x", "mm", "M_t",  "Nm"), 
                        ("Bending moment [z]",  "x",  "mm",  "M_{b,z}",  "Nm"), 
                        ("Bending moment [y]",  "x",  "mm",  "M_{b,y}",  "Nm"))
    wstrings = (("",  "",  "",  "",  ""), 
                        ("Translation [y]",  "x",  "mm",  "w_y",  "mm"), 
                        ("Translation [z]",  "x",  "mm",  "w_z",  "mm"))
    sigmaNstrings =  (("Normal stress [x]",  "x", "mm", "\sigma_x",  u"N/mm²"), 
                        ("Shear stress [y]",  "x",  "mm",  "\sigma_y",  u"N/mm²"), 
                        ("Shear stress [z]",  "x",  "mm",  "\sigma_z",  u"N/mm²"))
    sigmaBstrings = (("Torque stress [x]",  "x", "mm", "\tau_t",  u"N/mm²"), 
                        ("Bending stress [z]",  "x",  "mm",  "\sigma_{b,z}",  u"N/mm²"), 
                        ("Bending stress [y]",  "x",  "mm",  "\sigma_{b,y}",  u"N/mm²"))

    def __init__(self, parent):
        self.parent = parent
        self.doc = parent.doc
        self.feature = ShaftFeature(self.doc)
        # List of shaft segments (each segment has a different diameter)
        self.segments = []
        # The diagrams
        self.diagrams = {} # map of function name against Diagram object
        # Calculation of shaft
        self.F = [None,  None,  None] # force in direction of [x,y,z]-axis
        self.M = [None,  None,  None] # bending moment around [x,z,y]-axis
        self.w = [None,  None,  None] # Shaft translation due to bending
        self.sigmaN = [None,  None,  None]  # normal stress in direction of x-axis, shear stress in direction of [y,z]-axis
        self.sigmaB = [None,  None,  None]  # # torque stress around x-axis, maximum bending stress in direction of [y,z]-axis

    def getLengthTo(self, index):
        "Get the total length of all segments up to the given one"
        result = 0.0
        for i in range(index):
            result += self.segments[i].length
        return result

    def addSegment(self, l, d, di):
        self.segments.append(ShaftSegment(l,d,di))
        self.feature.addSegment(l, d, di)
        # We don't call equilibrium() here because the new segment has no constraints defined yet
        # Fix face reference of fixed segment if it is the last one
        for i in range(1,  len(self.segments)):
            if self.segments[i].constraintType != "Fixed":
                continue
            if i == len(self.segments) - 1:
                self.segments[index].constraint.References = [( self.feature.feature,  "Face%u" % (2 * (index+1) + 1) )]
            else:
                # Remove reference since it is now in the middle of the shaft (which is not allowed)
                self.segments[index].constraint.References = [(None,  "")]           

    def updateSegment(self, index, length = None, diameter = None, innerdiameter = None):
        oldLength = self.segments[index].length
            
        if length is not None:
            self.segments[index].length = length
        if diameter is not None:
            self.segments[index].diameter = diameter
        if innerdiameter is not None:
            self.segments[index].innerdiameter = innerdiameter
        
        self.feature.updateSegment(index, oldLength, self.segments[index].length, self.segments[index].diameter, self.segments[index].innerdiameter)
        self.equilibrium()
        self.updateDiagrams()

    def updateConstraint(self, index, constraintType):
        if (constraintType is not None):
            # Did the constraint type change?
            if (self.segments[index].constraintType != "None") and (self.segments[index].constraintType != constraintType):
                self.doc.removeObject(self.segments[index].constraint.Name)
                self.segments[index].constraint = None
            
            self.segments[index].constraintType = constraintType   
            
            # Create constraint if it does not exist yet or has changed
            if self.segments[index].constraint is None:
                if (constraintType == "Force"):
                    # TODO: Create a reference point and put the force onto it
                    constraint = self.doc.addObject("Fem::ConstraintForce","ShaftConstraintForce") 
                    constraint.Force = 1000.0
                    self.segments[index].constraint = constraint
                elif (constraintType == "Fixed"):
                    # TODO: Use robust reference as soon as it is available for the face
                    constraint = self.doc.addObject("Fem::ConstraintFixed","ShaftConstraintFixed")
                    if index == 0:
                        constraint.References = [( self.feature.feature,  "Face1")]
                    elif index == len(self.segments) - 1:
                        constraint.References = [( self.feature.feature,  "Face%u" % (2 * (index+1) + 1) )]
                    self.segments[index].constraint = constraint
                elif (constraintType == "Bearing"):
                    # TODO: Use robust reference as soon as it is available for the cylindrical face reference
                    constraint = self.doc.addObject("Fem::ConstraintBearing","ShaftConstraintBearing")                    
                    constraint.References = [( self.feature.feature,  "Face%u" % (2 * (index+1)) )]         
                    constraint.AxialFree = True
                    self.segments[index].constraint = constraint                    
                elif (constraintType == "Pulley"):
                    constraint= self.doc.addObject("Fem::ConstraintPulley","ShaftConstraintPulley")
                    constraint.References = [( self.feature.feature,  "Face%u" % (2 * (index+1)) )]               
                    self.segments[index].constraint = constraint                    
                elif (constraintType == "Gear"):
                    constraint = self.doc.addObject("Fem::ConstraintGear","ShaftConstraintGear")       
                    constraint.References = [( self.feature.feature,  "Face%u" % (2 * (index+1)) )]               
                    self.segments[index].constraint = constraint                      
            
        self.equilibrium()
        self.updateDiagrams()
        
    def editConstraint(self,  index):
        if (self.segments[index].constraint is not None):
            FreeCADGui.activeDocument().setEdit(self.segments[index].constraint.Name)
        
    def getConstraint(self,  index):
        return self.segments[index].constraint

    def updateEdge(self, column, start):
        App.Console.PrintMessage("Not implemented yet - waiting for robust references...")
        return
        if self.sketchClosed is not True:
            return
        # Create a chamfer or fillet at the start or end edge of the segment
        if start is True:
            row = rowStartEdgeType
            idx = 0
        else:
            row = rowEndEdgeType
            idx = 1

        edgeType = self.tableWidget.item(row, column).text()[0].upper()
        if not ((edgeType == "C") or (edgeType == "F")):
            return # neither chamfer nor fillet defined

        if edgeType == "C":
            objName = self.doc.addObject("PartDesign::Chamfer","ChamferShaft%u" % (column * 2 + idx))
        else:
            objName = self.doc.addObject("PartDesign::Fillet","FilletShaft%u" % (column * 2 + idx))
        if objName == "":
            return

        edgeName = "Edge%u" % self.getEdgeIndex(column, idx, edgeType)
        self.doc.getObject(objName).Base = (self.doc.getObject("RevolutionShaft"),"[%s]" % edgeName)
        # etc. etc.

    def getEdgeIndex(self, column, startIdx):
        # FIXME: This is impossible without robust references anchored in the sketch!!!
        return

    def updateDiagrams(self):                            
        for ax in range(3):
            if self.F[ax] is not None:
                if self.F[ax].name in self.diagrams:
                    self.diagrams[self.F[ax].name].update(self.F[ax], self.getLengthTo(len(self.segments)) / 1000.0)
            if self.M[ax] is not None:
                if self.M[ax].name in self.diagrams:
                    self.diagrams[self.M[ax].name].update(self.M[ax], self.getLengthTo(len(self.segments)) / 1000.0)
            if self.w[ax] is not None:
                if self.w[ax].name in self.diagrams:
                    self.diagrams[self.w[ax].name].update(self.w[ax], self.getLengthTo(len(self.segments)) / 1000.0)
            if self.sigmaN[ax] is not None:
                if self.sigmaN[ax].name in self.diagrams:
                    self.diagrams[self.sigmaN[ax].name].update(self.sigmaN[ax], self.getLengthTo(len(self.segments)) / 1000.0)
            if self.sigmaB[ax] is not None:
                if self.sigmaB[ax].name in self.diagrams:
                    self.diagrams[self.sigmaB[ax].name].update(self.sigmaB[ax], self.getLengthTo(len(self.segments)) / 1000.0)

    def showDiagram(self,  which):                
        if which in self.Fstr:
            ax = self.Fstr.index(which)
            text = self.Qstrings[ax]
            if self.F[ax] is None:
                # No data
                return
            if self.F[ax].name in self.diagrams:
                # Diagram is already open, close it again
                self.diagrams[self.F[ax].name].close()
                del (self.diagrams[self.F[ax].name])
                return
            self.diagrams[self.F[ax].name] = Diagram()
            self.diagrams[self.F[ax].name].create(text[0], self.F[ax], self.getLengthTo(len(self.segments)) / 1000.0, text[1], text[2], 1000.0, text[3], text[4], 1.0, 10)
        elif which in self.Mstr:
            ax = self.Mstr.index(which)
            text = self.Mstrings[ax]
            if self.M[ax] is None:
                # No data
                return
            if self.M[ax].name in self.diagrams:
                # Diagram is already open, close it again
                self.diagrams[self.M[ax].name].close()
                del (self.diagrams[self.M[ax].name])
                return
            self.diagrams[self.M[ax].name] = Diagram()
            self.diagrams[self.M[ax].name].create(text[0], self.M[ax], self.getLengthTo(len(self.segments)) / 1000.0, text[1], text[2], 1000.0, text[3], text[4], 1.0, 20)
        elif which in self.wstr:
            ax = self.wstr.index(which)
            text = self.wstrings[ax]
            if self.w[ax] is None:
                # No data
                return
            if self.w[ax].name in self.diagrams:
                # Diagram is already open, close it again
                self.diagrams[self.w[ax].name].close()
                del (self.diagrams[self.w[ax].name])
                return
            self.diagrams[self.w[ax].name] = Diagram()
            self.diagrams[self.w[ax].name].create(text[0], self.w[ax], self.getLengthTo(len(self.segments)) / 1000.0, text[1], text[2], 1000.0, text[3], text[4], 1000.0, 30)
        elif which in self.sigmaNstr:
            ax = self.sigmaNstr.index(which)
            text = self.sigmaNstrings[ax]
            if self.sigmaN[ax] is None:
                # No data
                return
            if self.sigmaN[ax].name in self.diagrams:
                # Diagram is already open, close it again
                self.diagrams[self.sigmaN[ax].name].close()
                del (self.diagrams[self.sigmaN[ax].name])
                return
            self.diagrams[self.sigmaN[ax].name] = Diagram()
            self.diagrams[self.sigmaN[ax].name].create(text[0], self.sigmaN[ax], self.getLengthTo(len(self.segments)) / 1000.0, text[1], text[2], 1000.0, text[3], text[4], 1.0E-6, 10)
        elif which in self.sigmaBstr:
            ax = self.sigmaBstr.index(which)
            text = self.sigmaBstrings[ax]
            if self.sigmaB[ax] is None:
                # No data
                return
            if self.sigmaB[ax].name in self.diagrams:
                # Diagram is already open, close it again
                self.diagrams[self.sigmaB[ax].name].close()
                del (self.diagrams[self.sigmaB[ax].name])
                return
            self.diagrams[self.sigmaB[ax].name] = Diagram()
            self.diagrams[self.sigmaB[ax].name].create(text[0], self.sigmaB[ax], self.getLengthTo(len(self.segments)) / 1000.0, text[1], text[2], 1000.0, text[3], text[4], 1.0E-6, 20)
    
    def addTo(self,  dict,  location,  value):
        if location not in dict:
            dict[location] = value
        else:
            dict[location] += value
    
    def equilibrium(self):
        # Build equilibrium equations
        try:
            import numpy as np
        except ImportError:
            FreeCAD.Console.PrintMessage("numpy is not installed on your system\n")
            raise ImportError("numpy not installed")
                
        # Initialization of structures. All three axes are handled separately so everything is 3-fold
        # dictionaries of (location : outer force/moment) with reverse sign, which means that the segment functions for the section force and section moment
        # created from them will have signs as by the convention in 
        # http://www.umwelt-campus.de/ucb/fileadmin/users/90_t.preussler/dokumente/Skripte/TEMECH/TMI/Ebene_Balkenstatik.pdf (page 10)
        # (see also example on page 19)
        forces = [{0.0:0.0},  {0.0:0.0},  {0.0:0.0}] 
        moments = [{0.0:0.0},  {0.0:0.0},  {0.0:0.0}]
        # Boundary conditions for shaft bending line
        tangents = [[],  [],  []] # Tangents to shaft bending line
        translations = [[],  [],  []] # Shaft displacement
        # Variable names, e.g. Fx, Mz. Because the system must be exactly determined, not more than two independent variables for each 
        # force/moment per axis are possible (if there are more no solution is calculated)
        variableNames = [[""],  [""], [""]] 
        # # dictionary of (variableName : location) giving the x-coordinate at which the force/moment represented by the variable acts on the shaft
        locations = {}
        # Coefficients of the equilibrium equations in the form a = b * F1 + c * F2 and d = e * M1 + f * M2
        # LHS (variables a1, a2, a3, d3) initialized to zero
        coefficientsF = [[0], [0], [0]]
        coefficientsM = [[0], [0], [0]]

        for i in range(len(self.segments)):
            cType = self.segments[i].constraintType
            constraint = self.segments[i].constraint
            
            if cType == "Fixed":
                # Fixed segment
                if i == 0:
                    # At beginning of shaft
                    location = 0                    
                elif i == len(self.segments) - 1:
                    # At end of shaft
                    location = self.getLengthTo(len(self.segments)) / 1000.0 # convert to meters                    
                else:
                    # TODO: Better error message
                    FreeCAD.Console.PrintMessage("Fixed constraint must be at beginning or end of shaft\n")
                    return
                    
                for ax in range(3):        
                    # Create a new reaction force
                    variableNames[ax].append("%s%u" % (self.Fstr[ax], i))
                    coefficientsF[ax].append(1)
                    # Register location of reaction force             
                    locations["%s%u" % (self.Fstr[ax], i)] = location
                    # Boundary conditions for the translations
                    tangents[ax].append((location,  0.0))
                    translations[ax].append((location,  0.0))
                coefficientsM[0].append(0) # Reaction force contributes no moment around x axis
                coefficientsM[1].append(location) # Reaction force contributes a positive moment around z axis
                coefficientsM[2].append(-location) # Reaction force contributes a negative moment around y axis  
                
                for ax in range(3):
                    # Create a new reaction moment
                    variableNames[ax].append("%s%u" % (self.Mstr[ax], i))
                    coefficientsF[ax].append(0)
                    coefficientsM[ax].append(1)                                   
                    locations["%s%u" % (self.Mstr[ax], i)] = location
                    
            elif cType == "Force":
                # Static force (currently force on midpoint of segment only)
                force = constraint.DirectionVector.multiply(constraint.Force)                
                # TODO: Extract value of the location from geometry
                location = (self.getLengthTo(i) + self.segments[i].length/2.0)  / 1000.0
                # The force itself
                for ax in range(3): 
                    if abs(force[ax]) > 0.0:
                        coefficientsF[ax][0] = coefficientsF[ax][0] - force[ax] # neg. because this coefficient is on the LHS of the equilibrium equation
                        self.addTo(forces[ax],  location,  -force[ax]) # neg. to fulfill the convention mentioned above
                # Moments created by the force (by definition no moment is created by the force in x-direction)                
                if abs(force[1]) > 0.0:
                    coefficientsM[1][0] = coefficientsM[1][0] - force[1] * location # moment around z-axis
                    self.addTo(moments[1],  location,  0)
                if abs(force[2]) > 0.0:
                    coefficientsM[2][0] = coefficientsM[2][0] + force[2] * location # moment around y-axis
                    self.addTo(moments[2],  location,  0) # No outer moment acts here!
                    
            elif cType == "Bearing":
                 location = constraint.BasePoint.x / 1000.0 # TODO: This assumes that the shaft feature starts with the first segment at (0,0,0)  and its axis corresponds to the x-axis
                 # Bearing reaction forces. TODO: the bearing is assumed to not induce any reaction moments
                 start = (0 if constraint.AxialFree == False else 1)
                 for ax in range(start, 3):
                    variableNames[ax].append("%s%u" % (self.Fstr[ax], i))
                    coefficientsF[ax].append(1)
                    locations["%s%u" % (self.Fstr[ax], i)] = location
                    # Boundary condition
                    translations[ax].append((location,  0.0))
                 if constraint.AxialFree == False:
                    coefficientsM[0].append(0)             # Reaction force contributes no moment around x axis
                 coefficientsM[1].append(location) # Reaction force contributes a positive moment around z axis
                 coefficientsM[2].append(-location) # Reaction force contributes a negative moment around y axis                 
                
            elif cType == "Gear":
                force = constraint.DirectionVector.multiply(constraint.Force) 
                location = constraint.BasePoint.x / 1000.0
                lever = [0,  constraint.Diameter/2.0/1000.0 * math.sin(constraint.ForceAngle / 180.0 * math.pi),  
                                    constraint.Diameter/2.0 /1000.0* math.cos(constraint.ForceAngle / 180.0 * math.pi)]
                
                # Effect of the gear force
                for ax in range(3):
                    if abs(force[ax]) > 0.0:
                        # Effect of the force
                        coefficientsF[ax][0] = coefficientsF[ax][0] - force[ax]
                        self.addTo(forces[ax],  location,  -force[ax])
                # Moments created by the force (by definition no moment is created by the force in x-direction)                
                if abs(force[1]) > 0.0:
                    coefficientsM[1][0] = coefficientsM[1][0] - force[1] * location # moment around z-axis
                    self.addTo(moments[1],  location,  0)
                if abs(force[2]) > 0.0:
                    coefficientsM[2][0] = coefficientsM[2][0] + force[2] * location # moment around y-axis
                    self.addTo(moments[2],  location,  0) # No outer moment acts here!
                
                # Moments created by the force and lever
                if abs(force[0]) > 0.0:
                    momenty = force[0] * lever[2]
                    momentz = force[0] * lever[1]
                    coefficientsM[1][0] = coefficientsM[1][0] + momentz # moment around z-axis
                    self.addTo(moments[1],  location,  momentz)
                    coefficientsM[2][0] = coefficientsM[2][0] - momenty # moment around y-axis
                    self.addTo(moments[2],  location,  -momenty)         
                if abs(force[1]) > 0.0:
                    moment = force[1] * lever[2]
                    coefficientsM[0][0] = coefficientsM[0][0] + moment
                    self.addTo(moments[0],  location,  moment)
                if abs(force[2]) > 0.0:
                    moment = force[2] * lever[1]
                    coefficientsM[0][0] = coefficientsM[0][0] - moment
                    self.addTo(moments[0],  location,  -moment)
            elif cType == "Pulley":
                forceAngle1 = (constraint.ForceAngle + constraint.BeltAngle + 90.0) / 180.0 * math.pi
                forceAngle2 = (constraint.ForceAngle - constraint.BeltAngle + 90.0) / 180.0 * math.pi
                #FreeCAD.Console.PrintMessage("BeltForce1: %f, BeltForce2: %f\n" % (constraint.BeltForce1,  constraint.BeltForce2))
                #FreeCAD.Console.PrintMessage("Angle1: %f, Angle2: %f\n" % (forceAngle1,  forceAngle2))
                force = [0,  -constraint.BeltForce1 * math.sin(forceAngle1) - constraint.BeltForce2 * math.sin(forceAngle2), 
                                   constraint.BeltForce1 * math.cos(forceAngle1) + constraint.BeltForce2 * math.cos(forceAngle2)]
                location = constraint.BasePoint.x / 1000.0                
                
                # Effect of the pulley forces
                for ax in range(3):
                    if abs(force[ax]) > 0.0:
                        # Effect of the force
                        coefficientsF[ax][0] = coefficientsF[ax][0] - force[ax]
                        self.addTo(forces[ax],  location,  -force[ax])
                # Moments created by the force (by definition no moment is created by the force in x-direction)                
                if abs(force[1] ) > 0.0:
                    coefficientsM[1][0] = coefficientsM[1][0] - force[1] * location # moment around z-axis
                    self.addTo(moments[1],  location,  0)
                if abs(force[2]) > 0.0:
                    coefficientsM[2][0] = coefficientsM[2][0] + force[2] * location # moment around y-axis
                    self.addTo(moments[2],  location,  0) # No outer moment acts here!
                
                # Torque
                moment = constraint.Force * (1 if constraint.IsDriven is True else -1) 
                coefficientsM[0][0] = coefficientsM[0][0] + moment
                self.addTo(moments[0],  location,  moment)

        areas = [None,  None,  None]
        areamoments = [None,  None,  None]
        bendingmoments = [None,  None,  None]
        torquemoments = [None,  None,  None]
        
        for ax in range(3):
            FreeCAD.Console.PrintMessage("Axis: %u\n" %  ax) 
            self.printEquilibrium(variableNames[ax], coefficientsF[ax])
            self.printEquilibrium(variableNames[ax], coefficientsM[ax])

            if  len(coefficientsF[ax]) <= 1:
                # Note: coefficientsF and coefficientsM always have the same length
                FreeCAD.Console.PrintMessage("Matrix is singular, no solution possible\n")
                self.parent.updateButtons(ax,  False)
                continue
                
            # Handle special cases. Note that the code above should ensure that coefficientsF and coefficientsM always have same length
            solution = [None,  None]
            if len(coefficientsF[ax]) == 2:
                if coefficientsF[ax][1] != 0.0 and coefficientsF[ax][0] != 0.0:
                    solution[0] = coefficientsF[ax][0] / coefficientsF[ax][1]
                if coefficientsM[ax][1] != 0.0 and coefficientsM[ax][0] != 0.0:
                    solution[1] = coefficientsM[ax][0] / coefficientsM[ax][1]
                    if abs(solution[0] - solution[1]) < 1E9:
                        FreeCAD.Console.PrintMessage("System is statically undetermined. No solution possible.\n")
                        self.parent.updateButtons(ax,  False)
                        continue
            else:            
                # Build matrix and vector for linear algebra solving algorithm           
               # TODO: This could easily be done manually... there are only 2 variables and 6 coefficients
                A = np.array([coefficientsF[ax][1:], coefficientsM[ax][1:]])                           
                b = np.array([coefficientsF[ax][0], coefficientsM[ax][0]])
                try:
                    solution = np.linalg.solve(A, b) # A * solution = b
                except np.linalg.linalg.LinAlgError as e:
                    FreeCAD.Console.PrintMessage(e.message)
                    FreeCAD.Console.PrintMessage(". No solution possible.\n")
                    self.parent.updateButtons(ax,  False)
                    continue

            # Complete dictionary of forces and moments with the two reaction forces that were calculated
            for i in range(2):
                if solution[i] is None:
                    continue
                FreeCAD.Console.PrintMessage("Reaction force/moment: %s = %f\n" % (variableNames[ax][i+1],  solution[i]))
                if variableNames[ax][i+1][0] == "M":
                    moments[ax][locations[variableNames[ax][i+1]]] = -solution[i]
                else:
                    forces[ax][locations[variableNames[ax][i+1]]] = -solution[i]

            FreeCAD.Console.PrintMessage(forces[ax])
            FreeCAD.Console.PrintMessage("\n")
            FreeCAD.Console.PrintMessage(moments[ax])
            FreeCAD.Console.PrintMessage("\n")
            
            # Forces
            self.F[ax] = SegmentFunction(self.Fstr[ax])
            self.F[ax].buildFromDict("x", forces[ax])
            self.parent.updateButton(1,  ax,  not self.F[ax].isZero())
            self.F[ax].output()        
            # Moments
            if ax == 0:
                self.M[0] = SegmentFunction(self.Mstr[0])
                self.M[0].buildFromDict("x",  moments[0])
            elif ax == 1:
                self.M[1] = self.F[1].integrated().negate()
                self.M[1].name = self.Mstr[1]
                self.M[1].addSegments(moments[1]) # takes care of boundary conditions       
            elif ax == 2:
                self.M[2] = self.F[2].integrated()        
                self.M[2].name = self.Mstr[2]
                self.M[2].addSegments(moments[2]) # takes care of boundary conditions       
            self.parent.updateButton(2,  ax,  not self.M[ax].isZero())
            self.M[ax].output()
            
            # Areas and area moments            
            location = 0.0
            areas[ax] = IntervalFunction() # A [m²]
            areamoments[ax] = IntervalFunction() # I [m⁴]
            bendingmoments[ax] = IntervalFunction() # W_b [m³]
            torquemoments[ax] = IntervalFunction() # W_t [m³]           
            
            for i in range(len(self.segments)):
                od = self.segments[i].diameter/1000.0
                id = self.segments[i].innerdiameter/1000.0
                length = self.segments[i].length/1000.0
                areas[ax].addInterval(location, length,  math.pi/4.0 * (math.pow(od,  2.0) - math.pow(id,  2.0)))
                areamoment = math.pi/64.0 * (math.pow(od,  4.0) - math.pow(id,  4.0))
                areamoments[ax].addInterval(location,  length,  areamoment)
                bendingmoments[ax].addInterval(location,  length,  areamoment / (od / 2.0))
                torquemoments[ax].addInterval(location,  length,  2 * (areamoment / (od / 2.0)))
                location += length
            
            # Bending line
            if ax > 0:
                if len(tangents[ax])+ len(translations[ax]) == 2:
                    # TODO: Get Young's module from material type instead of using 210000 N/mm² = 2.1E12 N/m²
                    self.w[ax] = TranslationFunction(self.M[ax].negated(),  2.1E12,  areamoments[ax], tangents[ax], translations[ax])
                    self.w[ax].name= self.wstr[ax]
                    self.parent.updateButton(3,  ax,  not self.w[ax].isZero())
                else:
                    self.parent.updateButton(3,  ax,  False)                
            
            # Normal/shear stresses and torque/bending stresses
            self.sigmaN[ax] = StressFunction(self.F[ax],  areas[ax])
            self.sigmaN[ax].name = self.sigmaNstr[ax]
            self.parent.updateButton(4,  ax,  not self.sigmaN[ax].isZero())
            if ax == 0:
                self.sigmaB[ax] = StressFunction(self.M[ax] ,  torquemoments[ax])
            else:
                self.sigmaB[ax] = StressFunction(self.M[ax],  bendingmoments[ax])
            self.sigmaB[ax].name = self.sigmaBstr[ax]
            self.parent.updateButton(5,  ax,  not self.sigmaB[ax].isZero())

    def printEquilibrium(self, var, coeff):
        # Auxiliary method for debugging purposes        
        for i in range(len(var)):
            if i == 0:
                FreeCAD.Console.PrintMessage("%f = " % coeff[i])
            else:
                FreeCAD.Console.PrintMessage("%f * %s" % (coeff[i], var[i]))
            if (i < len(var) - 1) and (i != 0):
                FreeCAD.Console.PrintMessage(" + ")
        FreeCAD.Console.PrintMessage("\n")



# ***************************************************************************
# *   Copyright (c) 2009 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides support for importing and exporting SVG files.

It enables importing/exporting objects directly to/from the 3D document
but doesn't handle the SVG output from the Drawing and TechDraw modules.

Currently it only reads the following entities:
* paths, lines, circular arcs, rects, circles, ellipses, polygons, polylines.

Currently unsupported:
* use, image.
"""
## @package importSVG
#  \ingroup DRAFT
#  \brief SVG file importer and exporter

# Check code with
# flake8 --ignore=E226,E266,E401,W503

__title__ = "FreeCAD Draft Workbench - SVG importer/exporter"
__author__ = "Yorik van Havre, Sebastian Hoogen"
__url__ = "https://www.freecadweb.org"

# TODO:
# ignoring CDATA
# handle image element (external references and inline base64)
# debug Problem with 'Sans' font from Inkscape
# debug Problem with fill color
# implement inheriting fill style from group
# handle relative units

import math
import os
import re
import xml.sax

import FreeCAD
import Draft
import DraftVecUtils
from FreeCAD import Vector
from FreeCAD import Console as FCC
from draftutils.translate import translate

if FreeCAD.GuiUp:
    from PySide import QtGui
    import FreeCADGui
    gui = True
    try:
        draftui = FreeCADGui.draftToolBar
    except AttributeError:
        draftui = None
else:
    gui = False
    draftui = None

# Save the native open function to avoid collisions
if open.__module__ in ['__builtin__', 'io']:
    pythonopen = open

svgcolors = {
    'Pink': (255, 192, 203),
    'Blue': (0, 0, 255),
    'Honeydew': (240, 255, 240),
    'Purple': (128, 0, 128),
    'Fuchsia': (255, 0, 255),
    'LawnGreen': (124, 252, 0),
    'Amethyst': (153, 102, 204),
    'Crimson': (220, 20, 60),
    'White': (255, 255, 255),
    'NavajoWhite': (255, 222, 173),
    'Cornsilk': (255, 248, 220),
    'Bisque': (255, 228, 196),
    'PaleGreen': (152, 251, 152),
    'Brown': (165, 42, 42),
    'DarkTurquoise': (0, 206, 209),
    'DarkGreen': (0, 100, 0),
    'MediumOrchid': (186, 85, 211),
    'Chocolate': (210, 105, 30),
    'PapayaWhip': (255, 239, 213),
    'Olive': (128, 128, 0),
    'Silver': (192, 192, 192),
    'PeachPuff': (255, 218, 185),
    'Plum': (221, 160, 221),
    'DarkGoldenrod': (184, 134, 11),
    'SlateGrey': (112, 128, 144),
    'MintCream': (245, 255, 250),
    'CornflowerBlue': (100, 149, 237),
    'Gold': (255, 215, 0),
    'HotPink': (255, 105, 180),
    'DarkBlue': (0, 0, 139),
    'LimeGreen': (50, 205, 50),
    'DeepSkyBlue': (0, 191, 255),
    'DarkKhaki': (189, 183, 107),
    'LightGrey': (211, 211, 211),
    'Yellow': (255, 255, 0),
    'Gainsboro': (220, 220, 220),
    'MistyRose': (255, 228, 225),
    'SandyBrown': (244, 164, 96),
    'DeepPink': (255, 20, 147),
    'Magenta': (255, 0, 255),
    'AliceBlue': (240, 248, 255),
    'DarkCyan': (0, 139, 139),
    'DarkSlateGrey': (47, 79, 79),
    'GreenYellow': (173, 255, 47),
    'DarkOrchid': (153, 50, 204),
    'OliveDrab': (107, 142, 35),
    'Chartreuse': (127, 255, 0),
    'Peru': (205, 133, 63),
    'Orange': (255, 165, 0),
    'Red': (255, 0, 0),
    'Wheat': (245, 222, 179),
    'LightCyan': (224, 255, 255),
    'LightSeaGreen': (32, 178, 170),
    'BlueViolet': (138, 43, 226),
    'LightSlateGrey': (119, 136, 153),
    'Cyan': (0, 255, 255),
    'MediumPurple': (147, 112, 219),
    'MidnightBlue': (25, 25, 112),
    'FireBrick': (178, 34, 34),
    'PaleTurquoise': (175, 238, 238),
    'PaleGoldenrod': (238, 232, 170),
    'Gray': (128, 128, 128),
    'MediumSeaGreen': (60, 179, 113),
    'Moccasin': (255, 228, 181),
    'Ivory': (255, 255, 240),
    'DarkSlateBlue': (72, 61, 139),
    'Beige': (245, 245, 220),
    'Green': (0, 128, 0),
    'SlateBlue': (106, 90, 205),
    'Teal': (0, 128, 128),
    'Azure': (240, 255, 255),
    'LightSteelBlue': (176, 196, 222),
    'DimGrey': (105, 105, 105),
    'Tan': (210, 180, 140),
    'AntiqueWhite': (250, 235, 215),
    'SkyBlue': (135, 206, 235),
    'GhostWhite': (248, 248, 255),
    'MediumTurquoise': (72, 209, 204),
    'FloralWhite': (255, 250, 240),
    'LavenderBlush': (255, 240, 245),
    'SeaGreen': (46, 139, 87),
    'Lavender': (230, 230, 250),
    'BlanchedAlmond': (255, 235, 205),
    'DarkOliveGreen': (85, 107, 47),
    'DarkSeaGreen': (143, 188, 143),
    'SpringGreen': (0, 255, 127),
    'Navy': (0, 0, 128),
    'Orchid': (218, 112, 214),
    'SaddleBrown': (139, 69, 19),
    'IndianRed': (205, 92, 92),
    'Snow': (255, 250, 250),
    'SteelBlue': (70, 130, 180),
    'MediumSlateBlue': (123, 104, 238),
    'Black': (0, 0, 0),
    'LightBlue': (173, 216, 230),
    'Turquoise': (64, 224, 208),
    'MediumVioletRed': (199, 21, 133),
    'DarkViolet': (148, 0, 211),
    'DarkGray': (169, 169, 169),
    'Salmon': (250, 128, 114),
    'DarkMagenta': (139, 0, 139),
    'Tomato': (255, 99, 71),
    'WhiteSmoke': (245, 245, 245),
    'Goldenrod': (218, 165, 32),
    'MediumSpringGreen': (0, 250, 154),
    'DodgerBlue': (30, 144, 255),
    'Aqua': (0, 255, 255),
    'ForestGreen': (34, 139, 34),
    'LemonChiffon': (255, 250, 205),
    'LightSlateGray': (119, 136, 153),
    'SlateGray': (112, 128, 144),
    'LightGray': (211, 211, 211),
    'Indigo': (75, 0, 130),
    'CadetBlue': (95, 158, 160),
    'LightYellow': (255, 255, 224),
    'DarkOrange': (255, 140, 0),
    'PowderBlue': (176, 224, 230),
    'RoyalBlue': (65, 105, 225),
    'Sienna': (160, 82, 45),
    'Thistle': (216, 191, 216),
    'Lime': (0, 255, 0),
    'Seashell': (255, 245, 238),
    'DarkRed': (139, 0, 0),
    'LightSkyBlue': (135, 206, 250),
    'YellowGreen': (154, 205, 50),
    'Aquamarine': (127, 255, 212),
    'LightCoral': (240, 128, 128),
    'DarkSlateGray': (47, 79, 79),
    'Khaki': (240, 230, 140),
    'DarkGrey': (169, 169, 169),
    'BurlyWood': (222, 184, 135),
    'LightGoldenrodYellow': (250, 250, 210),
    'MediumBlue': (0, 0, 205),
    'DarkSalmon': (233, 150, 122),
    'RosyBrown': (188, 143, 143),
    'LightSalmon': (255, 160, 122),
    'PaleVioletRed': (219, 112, 147),
    'Coral': (255, 127, 80),
    'Violet': (238, 130, 238),
    'Grey': (128, 128, 128),
    'LightGreen': (144, 238, 144),
    'Linen': (250, 240, 230),
    'OrangeRed': (255, 69, 0),
    'DimGray': (105, 105, 105),
    'Maroon': (128, 0, 0),
    'LightPink': (255, 182, 193),
    'MediumAquamarine': (102, 205, 170),
    'OldLace': (253, 245, 230)
}
svgcolorslower = \
    dict((key.lower(), value) for (key, value) in list(svgcolors.items()))


def getcolor(color):
    """Check if the given string is an RGB value, or if it is a named color.

    Parameters
    ----------
    color : str
        Color in hexadecimal format, long '#12ab9f' or short '#1af'

    Returns
    -------
    tuple
    (r, g, b, a)
        RGBA float tuple, where each value is between 0.0 and 1.0.
    """
    if color[0] == "#":
        # Color string '#12ab9f'
        if len(color) == 7:
            r = float(int(color[1:3], 16) / 255.0)
            g = float(int(color[3:5], 16) / 255.0)
            b = float(int(color[5:], 16) / 255.0)
        # Color string '#1af'
        elif len(color) == 4:
            # Expand the hex digits
            r = float(int(color[1], 16) * 17 / 255.0)
            g = float(int(color[2], 16) * 17 / 255.0)
            b = float(int(color[3], 16) * 17 / 255.0)
        return (r, g, b, 0.0)
    # Color string 'rgb(0.12,0.23,0.3,0.0)'
    elif color.lower().startswith('rgb('):
        cvalues = color[3:].lstrip('(').rstrip(')').replace('%', '').split(',')
        if '%' in color:
            r, g, b = [int(float(cv)) / 100.0 for cv in cvalues]
        else:
            r, g, b = [int(float(cv)) / 255.0 for cv in cvalues]
        return (r, g, b, 0.0)
    # Color string 'MediumAquamarine'
    else:
        v = svgcolorslower.get(color.lower())
        if v:
            r, g, b = [float(vf) / 255.0 for vf in v]
            return (r, g, b, 0.0)
        # for k, v in svgcolors.items():
        #    if k.lower() == color.lower():
        #        pass


def transformCopyShape(shape, m):
    """Apply transformation matrix m on given shape.

    Since OCCT 6.8.0 transformShape can be used to apply certain
    non-orthogonal transformations on shapes. This way a conversion
    to BSplines in transformGeometry can be avoided.

    @sa: Part::TopoShape::transformGeometry(), TopoShapePy::transformGeometry()
    @sa: Part::TopoShape::transformShape(), TopoShapePy::transformShape()

    Parameters
    ----------
    shape : Part::TopoShape
        A given shape
    m : Base::Matrix4D
        A transformation matrix

    Returns
    -------
    shape : Part::TopoShape
        The shape transformed by the matrix
    """
    # If there is no shear, these matrix operations will be very small
    _s1 = abs(m.A11**2 + m.A12**2 - m.A21**2 - m.A22**2)
    _s2 = abs(m.A11 * m.A21 + m.A12 * m.A22)
    if _s1 < 1e-8 and _s2 < 1e-8:
        try:
            newshape = shape.copy()
            newshape.transformShape(m)
            return newshape
        # Older versions of OCCT will refuse to work on
        # non-orthogonal matrices
        except Part.OCCError:
            pass
    return shape.transformGeometry(m)


def getsize(length, mode='discard', base=1):
    """Parse the length string containing number and unit.

    Parameters
    ----------
    length : str
        The length is a string, including sign, exponential notation,
        and unit: '+56215.14565E+6mm', '-23.156e-2px'.
    mode : str, optional
        One of 'discard', 'tuple', 'css90.0', 'css96.0', 'mm90.0', 'mm96.0'.
        'discard' (default), it discards the unit suffix, and extracts
            a number from the given string.
        'tuple', return number and unit as a tuple
        'css90.0', convert the unit to pixels assuming 90 dpi
        'css96.0', convert the unit to pixels assuming 96 dpi
        'mm90.0', convert the unit to millimeters assuming 90 dpi
        'mm96.0', convert the unit to millimeters assuming 96 dpi
    base : float, optional
        A base to scale the length.

    Returns
    -------
    float
        The numeric value of the length, as is, or transformed to
        millimeters or pixels.
    float, string
        A tuple with the numeric value, and the unit if `mode='tuple'`.
    """
    # Dictionaries to convert units to millimeters or pixels.
    #
    # The `em` and `ex` units are typographical units used in systems
    # like LaTeX. Here the conversion factors are arbitrarily chosen,
    # as they should depend on a specific font size used.
    #
    # The percentage factor is arbitrarily chosen, as it should depend
    # on the viewport size or for filling patterns on the bounding box.
    if mode == 'mm90.0':
        tomm = {
            '': 25.4/90,  # default
            'px': 25.4/90,
            'pt': 4.0/3 * 25.4/90,
            'pc': 15 * 25.4/90,
            'mm': 1.0,
            'cm': 10.0,
            'in': 25.4,
            'em': 15 * 2.54/90,
            'ex': 10 * 2.54/90,
            '%': 100
        }
    elif mode == 'mm96.0':
        tomm = {
            '': 25.4/96,  # default
            'px': 25.4/96,
            'pt': 4.0/3 * 25.4/96,
            'pc': 15 * 25.4/96,
            'mm': 1.0,
            'cm': 10.0,
            'in': 25.4,
            'em': 15 * 2.54/96,
            'ex': 10 * 2.54/96,
            '%': 100
        }
    elif mode == 'css90.0':
        topx = {
            '': 1.0,  # default
            'px': 1.0,
            'pt': 4.0/3,
            'pc': 15,
            'mm': 90.0/25.4,
            'cm': 90.0/254.0,
            'in': 90,
            'em': 15,
            'ex': 10,
            '%': 100
        }
    elif mode == 'css96.0':
        topx = {
            '': 1.0,  # default
            'px': 1.0,
            'pt': 4.0/3,
            'pc': 15,
            'mm': 96.0/25.4,
            'cm': 96.0/254.0,
            'in': 96,
            'em': 15,
            'ex': 10,
            '%': 100
        }

    # Extract a number from a string like '+56215.14565E+6mm'
    _num = '([-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)'
    _unit = '(px|pt|pc|mm|cm|in|em|ex|%)?'
    _full_num = _num + _unit
    number, exponent, unit = re.findall(_full_num, length)[0]
    if mode == 'discard':
        return float(number)
    elif mode == 'tuple':
        return float(number), unit
    elif mode == 'isabsolute':
        return unit in ('mm', 'cm', 'in', 'px', 'pt')
    elif mode == 'mm96.0' or mode == 'mm90.0':
        return float(number) * tomm[unit]
    elif mode == 'css96.0' or mode == 'css90.0':
        if unit != '%':
            return float(number) * topx[unit]
        else:
            return float(number) * base


def makewire(path, checkclosed=False, donttry=False):
    '''Try to make a wire out of the list of edges.

    If the wire functions fail or the wire is not closed,
    if required the TopoShapeCompoundPy::connectEdgesToWires()
    function is used.

    Parameters
    ----------
    path : Part.Edge
        A collection of edges
    checkclosed : bool, optional
        Default is `False`.
    donttry : bool, optional
        Default is `False`. If it's `True` it won't try to check
        for a closed path.

    Returns
    -------
    Part::Wire
        A wire created from the ordered edges.
    Part::Compound
        A compound made of the edges, but unable to form a wire.
    '''
    if not donttry:
        try:
            import Part
            sh = Part.Wire(Part.__sortEdges__(path))
            # sh = Part.Wire(path)
            isok = (not checkclosed) or sh.isClosed()
            if len(sh.Edges) != len(path):
                isok = False
        # BRep_API: command not done
        except Part.OCCError:
            isok = False
    if donttry or not isok:
        # Code from wmayer forum p15549 to fix the tolerance problem
        # original tolerance = 0.00001
        comp = Part.Compound(path)
        _sh = comp.connectEdgesToWires(False,
                                       10**(-1 * (Draft.precision() - 2)))
        sh = _sh.Wires[0]
        if len(sh.Edges) != len(path):
            FCC.PrintWarning("Unable to form a wire\n")
            sh = comp
    return sh


def arccenter2end(center, rx, ry, angle1, angledelta, xrotation=0.0):
    '''Calculate start and end points, and flags of an arc.

    Calculate start and end points, and flags of an arc given in
    ``center parametrization``.
    See http://www.w3.org/TR/SVG/implnote.html#ArcImplementationNotes

    Parameters
    ----------
    center : Base::Vector3
        Coordinates of the center of the ellipse.
    rx : float
        Radius of the ellipse, semi-major axis in the X direction
    ry : float
        Radius of the ellipse, semi-minor axis in the Y direction
    angle1 : float
        Initial angle in radians
    angledelta : float
        Additional angle in radians
    xrotation : float, optional
        Default 0. Rotation around the Z axis

    Returns
    -------
    v1, v2, largerc, sweep
        Tuple indicating the end points of the arc, and two boolean values
        indicating whether the arc is less than 180 degrees or not,
        and whether the angledelta is negative.
    '''
    vr1 = Vector(rx * math.cos(angle1), ry * math.sin(angle1), 0)
    vr2 = Vector(rx * math.cos(angle1 + angledelta),
                 ry * math.sin(angle1 + angledelta),
                 0)
    mxrot = FreeCAD.Matrix()
    mxrot.rotateZ(xrotation)
    v1 = mxrot.multiply(vr1).add(center)
    v2 = mxrot.multiply(vr2).add(center)
    fa = ((abs(angledelta) / math.pi) % 2) > 1  # < 180 deg
    fs = angledelta < 0
    return v1, v2, fa, fs


def arcend2center(lastvec, currentvec, rx, ry,
                  xrotation=0.0, correction=False):
    '''Calculate the possible centers for an arc in endpoint parameterization.

    Calculate (positive and negative) possible centers for an arc given in
    ``endpoint parametrization``.
    See http://www.w3.org/TR/SVG/implnote.html#ArcImplementationNotes

    the sweepflag is interpreted as: sweepflag <==>  arc is travelled clockwise

    Parameters
    ----------
    lastvec : Base::Vector3
        First point of the arc.
    currentvec : Base::Vector3
        End point (current) of the arc.
    rx : float
        Radius of the ellipse, semi-major axis in the X direction.
    ry : float
        Radius of the ellipse, semi-minor axis in the Y direction.
    xrotation : float, optional
        Default is 0. Rotation around the Z axis, in radians (CCW).
    correction : bool, optional
        Default is `False`. If it is `True`, the radii will be scaled
        by a factor.

    Returns
    -------
    list, (float, float)
        A tuple that consists of one list, and a tuple of radii.
    [(positive), (negative)], (rx, ry)
        The first element of the list is the positive tuple,
        the second is the negative tuple.
    [(Base::Vector3, float, float),
    (Base::Vector3, float, float)], (float, float)
        Types
    [(vcenter+, angle1+, angledelta+),
    (vcenter-, angle1-, angledelta-)], (rx, ry)
        The first element of the list is the positive tuple,
        consisting of center, angle, and angle increment;
        the second element is the negative tuple.
    '''
    # scalefacsign = 1 if (largeflag != sweepflag) else -1
    rx = float(rx)
    ry = float(ry)
    v0 = lastvec.sub(currentvec)
    v0.multiply(0.5)
    m1 = FreeCAD.Matrix()
    m1.rotateZ(-xrotation)  # eq. 5.1
    v1 = m1.multiply(v0)
    if correction:
        eparam = v1.x**2 / rx**2 + v1.y**2 / ry**2
        if eparam > 1:
            eproot = math.sqrt(eparam)
            rx = eproot * rx
            ry = eproot * ry
    denom = rx**2 * v1.y**2 + ry**2 * v1.x**2
    numer = rx**2 * ry**2 - denom
    results = []

    # If the division is very small, set the scaling factor to zero,
    # otherwise try to calculate it by taking the square root
    if abs(numer/denom) < 10**(-1 * (Draft.precision())):
        scalefacpos = 0
    else:
        try:
            scalefacpos = math.sqrt(numer/denom)
        except ValueError:
            FCC.PrintMessage('sqrt(%f/%f)\n' % (numer, denom))
            scalefacpos = 0
    # Calculate two values because the square root may be positive or negative
    for scalefacsign in (1, -1):
        scalefac = scalefacpos * scalefacsign
        # Step2 eq. 5.2
        vcx1 = Vector(v1.y * rx/ry, -v1.x * ry/rx, 0).multiply(scalefac)
        m2 = FreeCAD.Matrix()
        m2.rotateZ(xrotation)
        centeroff = currentvec.add(lastvec)
        centeroff.multiply(0.5)
        vcenter = m2.multiply(vcx1).add(centeroff)  # Step3 eq. 5.3
        # angle1 = Vector(1, 0, 0).getAngle(Vector((v1.x - vcx1.x)/rx,
        #                                          (v1.y - vcx1.y)/ry,
        #                                          0))  # eq. 5.5
        # angledelta = Vector((v1.x - vcx1.x)/rx,
        #                     (v1.y - vcx1.y)/ry,
        #                     0).getAngle(Vector((-v1.x - vcx1.x)/rx,
        #                                        (-v1.y - vcx1.y)/ry,
        #                                        0))  # eq. 5.6
        # we need the right sign for the angle
        angle1 = DraftVecUtils.angle(Vector(1, 0, 0),
                                     Vector((v1.x - vcx1.x)/rx,
                                            (v1.y - vcx1.y)/ry,
                                            0))  # eq. 5.5
        angledelta = DraftVecUtils.angle(Vector((v1.x - vcx1.x)/rx,
                                                (v1.y - vcx1.y)/ry,
                                                0),
                                         Vector((-v1.x - vcx1.x)/rx,
                                                (-v1.y - vcx1.y)/ry,
                                                0))  # eq. 5.6
        results.append((vcenter, angle1, angledelta))
    return results, (rx, ry)


def getrgb(color):
    """Return an RGB hexadecimal string '#00aaff' from a FreeCAD color.

    Parameters
    ----------
    color : App::Color::Color
        FreeCAD color.

    Returns
    -------
    str
        The hexadecimal string representation of the color '#00aaff'.
    """
    r = str(hex(int(color[0] * 255)))[2:].zfill(2)
    g = str(hex(int(color[1] * 255)))[2:].zfill(2)
    b = str(hex(int(color[2] * 255)))[2:].zfill(2)
    return "#" + r + g + b


class svgHandler(xml.sax.ContentHandler):
    """Parse SVG files and create FreeCAD objects."""

    def __init__(self):
        """Retrieve Draft parameters and initialize."""
        _prefs = "User parameter:BaseApp/Preferences/Mod/Draft"
        params = FreeCAD.ParamGet(_prefs)
        self.style = params.GetInt("svgstyle")
        self.disableUnitScaling = params.GetBool("svgDisableUnitScaling",
                                                 False)
        self.count = 0
        self.transform = None
        self.grouptransform = []
        self.lastdim = None
        self.viewbox = None
        self.symbols = {}
        self.currentsymbol = None
        self.svgdpi = 1.0

        global Part
        import Part

        if gui and draftui:
            r = float(draftui.color.red()/255.0)
            g = float(draftui.color.green()/255.0)
            b = float(draftui.color.blue()/255.0)
            self.lw = float(draftui.linewidth)
        else:
            self.lw = float(params.GetInt("linewidth"))
            c = params.GetUnsigned("color")
            r = float(((c >> 24) & 0xFF)/255)
            g = float(((c >> 16) & 0xFF)/255)
            b = float(((c >> 8) & 0xFF)/255)
        self.col = (r, g, b, 0.0)

    def format(self, obj):
        """Apply styles to the object if the graphical interface is up."""
        if FreeCAD.GuiUp:
            v = obj.ViewObject
            if self.color:
                v.LineColor = self.color
            if self.width:
                v.LineWidth = self.width
            if self.fill:
                v.ShapeColor = self.fill

    def startElement(self, name, attrs):
        """Re-organize data into a nice clean dictionary.

        Parameters
        ----------
        name : str
            Name of the element: 'path', 'rect', 'line', 'polyline',
            'polygon', 'ellipse', 'circle', 'text', 'tspan', 'symbol'
        attrs : iterable
            Dictionary of content of the elements
        """
        self.count += 1
        FCC.PrintMessage('processing element %d: %s\n' % (self.count, name))
        FCC.PrintMessage('existing group transform: %s\n'
                         % (str(self.grouptransform)))

        data = {}
        for (keyword, content) in list(attrs.items()):
            # print(keyword, content)
            if keyword != "style":
                content = content.replace(',', ' ')
                content = content.split()
            # print(keyword, content)
            data[keyword] = content

        # If it's the first element, which is <svg>,
        # check if the file is created by Inkscape, and its version,
        # in order to consider some attributes of the SVG file.
        if self.count == 1 and name == 'svg':
            if 'inkscape:version' in data:
                inks_doc_name = attrs.getValue('sodipodi:docname')
                inks_full_ver = attrs.getValue('inkscape:version')
                inks_ver_pars = re.search("\d+\.\d+", inks_full_ver)
                if inks_ver_pars != None:
                    inks_ver_f = float(inks_ver_pars.group(0))
                else:
                    inks_ver_f = 99.99
                # Inkscape before 0.92 used 90 dpi as resolution
                # Newer versions use 96 dpi
                if inks_ver_f < 0.92:
                    self.svgdpi = 90.0
                else:
                    self.svgdpi = 96.0
            if 'inkscape:version' not in data:
                _msg = ("This SVG file does not appear to have been produced "
                        "by Inkscape. If it does not contain absolute units "
                        "then a DPI setting will be used.")
                _qst = ("Do you wish to use 96 dpi? Choosing 'No' "
                        "will use the older standard 90 dpi.")
                if FreeCAD.GuiUp:
                    msgBox = QtGui.QMessageBox()
                    msgBox.setText(translate("ImportSVG", _msg))
                    msgBox.setInformativeText(translate("ImportSVG", _qst))
                    msgBox.setStandardButtons(QtGui.QMessageBox.Yes
                                              | QtGui.QMessageBox.No)
                    msgBox.setDefaultButton(QtGui.QMessageBox.No)
                    ret = msgBox.exec_()
                    if ret == QtGui.QMessageBox.Yes:
                        self.svgdpi = 96.0
                    else:
                        self.svgdpi = 90.0
                    if ret:
                        FCC.PrintMessage(translate("ImportSVG", _msg) + "\n")
                        FCC.PrintMessage(translate("ImportSVG", _qst) + "\n")
                        FCC.PrintMessage("*** User specified "
                                         + str(self.svgdpi) + " dpi ***\n")
                else:
                    self.svgdpi = 96.0
                    FCC.PrintMessage(_msg + "\n")
                    FCC.PrintMessage("*** Assuming " + str(self.svgdpi)
                                     + " dpi ***\n")
            if self.svgdpi == 1.0:
                FCC.PrintWarning("This SVG file (" + inks_doc_name + ") "
                                 "has an unrecognised format which means "
                                 "the dpi could not be determined; "
                                 "assuming 96 dpi\n")
                self.svgdpi = 96.0

        if 'style' in data:
            if not data['style']:
                # Empty style attribute stops inheriting from parent
                pass
            else:
                content = data['style'].replace(' ', '')
                content = content.split(';')
                for i in content:
                    pair = i.split(':')
                    if len(pair) > 1:
                        data[pair[0]] = pair[1]

        for k in ['x', 'y', 'x1', 'y1', 'x2', 'y2',
                  'r', 'rx', 'ry', 'cx', 'cy', 'width', 'height']:
            if k in data:
                data[k] = getsize(data[k][0], 'css' + str(self.svgdpi))

        for k in ['fill', 'stroke', 'stroke-width', 'font-size']:
            if k in data:
                if isinstance(data[k], list):
                    if data[k][0].lower().startswith("rgb("):
                        data[k] = ",".join(data[k])
                    else:
                        data[k] = data[k][0]

        # Extract style info
        self.fill = None
        self.color = None
        self.width = None
        self.text = None

        if name == 'svg':
            m = FreeCAD.Matrix()
            if not self.disableUnitScaling:
                if 'width' in data \
                        and 'height' in data \
                        and 'viewBox' in data:
                    vbw = float(data['viewBox'][2])
                    vbh = float(data['viewBox'][3])
                    w = attrs.getValue('width')
                    h = attrs.getValue('height')
                    self.viewbox = (vbw, vbh)
                    if len(self.grouptransform) == 0:
                        unitmode = 'mm' + str(self.svgdpi)
                    else:
                        # nested svg element
                        unitmode = 'css' + str(self.svgdpi)
                    abw = getsize(w, unitmode)
                    abh = getsize(h, unitmode)
                    sx = abw / vbw
                    sy = abh / vbh
                    _data = data.get('preserveAspectRatio', [])
                    preservearstr = ' '.join(_data).lower()
                    uniformscaling = round(sx/sy, 5) == 1
                    if uniformscaling:
                        m.scale(Vector(sx, sy, 1))
                    else:
                        FCC.PrintWarning('Scaling factors do not match!\n')
                        if preservearstr.startswith('none'):
                            m.scale(Vector(sx, sy, 1))
                        else:
                            # preserve the aspect ratio
                            if preservearstr.endswith('slice'):
                                sxy = max(sx, sy)
                            else:
                                sxy = min(sx, sy)
                            m.scale(Vector(sxy, sxy, 1))
                elif len(self.grouptransform) == 0:
                    # fallback to current dpi
                    m.scale(Vector(25.4/self.svgdpi, 25.4/self.svgdpi, 1))
            self.grouptransform.append(m)
        if 'fill' in data:
            if data['fill'][0] != 'none':
                self.fill = getcolor(data['fill'])
        if 'stroke' in data:
            if data['stroke'][0] != 'none':
                self.color = getcolor(data['stroke'])
        if 'stroke-width' in data:
            if data['stroke-width'] != 'none':
                self.width = getsize(data['stroke-width'],
                                     'css' + str(self.svgdpi))
        if 'transform' in data:
            m = self.getMatrix(attrs.getValue('transform'))
            if name == "g":
                self.grouptransform.append(m)
            else:
                self.transform = m
        else:
            if name == "g":
                self.grouptransform.append(FreeCAD.Matrix())

        if self.style == 1:
            self.color = self.col
            self.width = self.lw

        pathname = None
        if 'id' in data:
            pathname = data['id'][0]
            FCC.PrintMessage('name: %s\n' % pathname)

        # Process paths
        if name == "path":
            FCC.PrintMessage('data: %s\n' % str(data))

            if not pathname:
                pathname = 'Path'

            path = []
            point = []
            lastvec = Vector(0, 0, 0)
            lastpole = None
            # command = None
            relative = False
            firstvec = None

            if "freecad:basepoint1" in data:
                p1 = data["freecad:basepoint1"]
                p1 = Vector(float(p1[0]), -float(p1[1]), 0)
                p2 = data["freecad:basepoint2"]
                p2 = Vector(float(p2[0]), -float(p2[1]), 0)
                p3 = data["freecad:dimpoint"]
                p3 = Vector(float(p3[0]), -float(p3[1]), 0)
                obj = Draft.makeDimension(p1, p2, p3)
                self.applyTrans(obj)
                self.format(obj)
                self.lastdim = obj
                data['d'] = []

            _op = '([mMlLhHvVaAcCqQsStTzZ])'
            _op2 = '([^mMlLhHvVaAcCqQsStTzZ]*)'
            _command = '\s*?' + _op + '\s*?' + _op2 + '\s*?'
            pathcommandsre = re.compile(_command, re.DOTALL)

            _num = '[-+]?[0-9]*\.?[0-9]+'
            _exp = '([eE][-+]?[0-9]+)?'
            _point = '(' + _num + _exp + ')'
            pointsre = re.compile(_point, re.DOTALL)
            _commands = pathcommandsre.findall(' '.join(data['d']))
            for d, pointsstr in _commands:
                relative = d.islower()
                _points = pointsre.findall(pointsstr.replace(',', ' '))
                pointlist = [float(number) for number, exponent in _points]

                if (d == "M" or d == "m"):
                    x = pointlist.pop(0)
                    y = pointlist.pop(0)
                    if path:
                        # sh = Part.Wire(path)
                        sh = makewire(path)
                        if self.fill and sh.isClosed():
                            sh = Part.Face(sh)
                        sh = self.applyTrans(sh)
                        obj = self.doc.addObject("Part::Feature", pathname)
                        obj.Shape = sh
                        self.format(obj)
                        if self.currentsymbol:
                            self.symbols[self.currentsymbol].append(obj)
                        path = []
                        # if firstvec:
                        #    Move relative to last move command
                        #    not last draw command
                        #    lastvec = firstvec
                    if relative:
                        lastvec = lastvec.add(Vector(x, -y, 0))
                    else:
                        lastvec = Vector(x, -y, 0)
                    firstvec = lastvec
                    FCC.PrintMessage('move %s\n' % str(lastvec))
                    lastpole = None

                if (d == "L" or d == "l") \
                        or ((d == 'm' or d == 'M') and pointlist):
                    for x, y in zip(pointlist[0::2], pointlist[1::2]):
                        if relative:
                            currentvec = lastvec.add(Vector(x, -y, 0))
                        else:
                            currentvec = Vector(x, -y, 0)
                        if not DraftVecUtils.equals(lastvec, currentvec):
                            _seg = Part.LineSegment(lastvec, currentvec)
                            seg = _seg.toShape()
                            FCC.PrintMessage("line %s %s\n"
                                             % (lastvec, currentvec))
                            lastvec = currentvec
                            path.append(seg)
                        lastpole = None
                elif (d == "H" or d == "h"):
                    for x in pointlist:
                        if relative:
                            currentvec = lastvec.add(Vector(x, 0, 0))
                        else:
                            currentvec = Vector(x, lastvec.y, 0)
                        seg = Part.LineSegment(lastvec, currentvec).toShape()
                        lastvec = currentvec
                        lastpole = None
                        path.append(seg)
                elif (d == "V" or d == "v"):
                    for y in pointlist:
                        if relative:
                            currentvec = lastvec.add(Vector(0, -y, 0))
                        else:
                            currentvec = Vector(lastvec.x, -y, 0)
                        if lastvec != currentvec:
                            _seg = Part.LineSegment(lastvec, currentvec)
                            seg = _seg.toShape()
                            lastvec = currentvec
                            lastpole = None
                            path.append(seg)
                elif (d == "A" or d == "a"):
                    piter = zip(pointlist[0::7], pointlist[1::7],
                                pointlist[2::7], pointlist[3::7],
                                pointlist[4::7], pointlist[5::7],
                                pointlist[6::7])
                    for (rx, ry, xrotation,
                         largeflag, sweepflag,
                         x, y) in piter:
                        # support for large-arc and x-rotation is missing
                        if relative:
                            currentvec = lastvec.add(Vector(x, -y, 0))
                        else:
                            currentvec = Vector(x, -y, 0)
                        chord = currentvec.sub(lastvec)
                        # small circular arc
                        _precision = 10**(-1*Draft.precision())
                        if (not largeflag) and abs(rx - ry) < _precision:
                            # perp = chord.cross(Vector(0, 0, -1))
                            # here is a better way to find the perpendicular
                            if sweepflag == 1:
                                # clockwise
                                perp = DraftVecUtils.rotate2D(chord,
                                                              -math.pi/2)
                            else:
                                # anticlockwise
                                perp = DraftVecUtils.rotate2D(chord, math.pi/2)
                            chord.multiply(0.5)
                            if chord.Length > rx:
                                a = 0
                            else:
                                a = math.sqrt(rx**2 - chord.Length**2)
                            s = rx - a
                            perp.multiply(s/perp.Length)
                            midpoint = lastvec.add(chord.add(perp))
                            _seg = Part.Arc(lastvec, midpoint, currentvec)
                            seg = _seg.toShape()
                        # big arc or elliptical arc
                        else:
                            # Calculate the possible centers for an arc
                            # in 'endpoint parameterization'.
                            _xrot = math.radians(-xrotation)
                            (solution,
                             (rx, ry)) = arcend2center(lastvec,
                                                       currentvec,
                                                       rx, ry,
                                                       xrotation=_xrot,
                                                       correction=True)
                            # Chose one of the two solutions
                            negsol = (largeflag != sweepflag)
                            vcenter, angle1, angledelta = solution[negsol]
                            # print(angle1)
                            # print(angledelta)
                            if ry > rx:
                                rx, ry = ry, rx
                                swapaxis = True
                            else:
                                swapaxis = False
                            # print('Elliptical arc %s rx=%f ry=%f'
                            #       % (vcenter, rx, ry))
                            e1 = Part.Ellipse(vcenter, rx, ry)
                            if sweepflag:
                                # Step4
                                # angledelta = -(-angledelta % (2*math.pi))
                                # angledelta = (-angledelta % (2*math.pi))
                                angle1 = angle1 + angledelta
                                angledelta = -angledelta
                                # angle1 = math.pi - angle1

                            d90 = math.radians(90)
                            e1a = Part.Arc(e1,
                                           angle1 - swapaxis * d90,
                                           angle1 + angledelta
                                                  - swapaxis * d90)
                            # e1a = Part.Arc(e1,
                            #                angle1 - 0 * swapaxis * d90,
                            #                angle1 + angledelta
                            #                       - 0 * swapaxis * d90)
                            _precision = 10**(-1*Draft.precision())
                            if swapaxis or xrotation > _precision:
                                m3 = FreeCAD.Matrix()
                                m3.move(vcenter)
                                # 90
                                rot90 = FreeCAD.Matrix(0, -1, 0, 0, 1, 0)
                                # swapaxism = FreeCAD.Matrix(0, 1, 0, 0, 1, 0)
                                if swapaxis:
                                    m3 = m3.multiply(rot90)
                                m3.rotateZ(math.radians(-xrotation))
                                m3.move(vcenter.multiply(-1))
                                e1a.transform(m3)
                            seg = e1a.toShape()
                            if sweepflag:
                                seg.reverse()
                                # DEBUG
                                # obj = self.doc.addObject("Part::Feature",
                                #                       'DEBUG %s' % pathname)
                                # obj.Shape = seg
                                # _seg = Part.LineSegment(lastvec, currentvec)
                                # seg = _seg.toShape()
                        lastvec = currentvec
                        lastpole = None
                        path.append(seg)
                elif (d == "C" or d == "c") or (d == "S" or d == "s"):
                    smooth = (d == 'S' or d == 's')
                    if smooth:
                        piter = list(zip(pointlist[2::4],
                                         pointlist[3::4],
                                         pointlist[0::4],
                                         pointlist[1::4],
                                         pointlist[2::4],
                                         pointlist[3::4]))
                    else:
                        piter = list(zip(pointlist[0::6],
                                         pointlist[1::6],
                                         pointlist[2::6],
                                         pointlist[3::6],
                                         pointlist[4::6],
                                         pointlist[5::6]))
                    for p1x, p1y, p2x, p2y, x, y in piter:
                        if smooth:
                            if lastpole is not None and lastpole[0] == 'cubic':
                                pole1 = lastvec.sub(lastpole[1]).add(lastvec)
                            else:
                                pole1 = lastvec
                        else:
                            if relative:
                                pole1 = lastvec.add(Vector(p1x, -p1y, 0))
                            else:
                                pole1 = Vector(p1x, -p1y, 0)
                        if relative:
                            currentvec = lastvec.add(Vector(x, -y, 0))
                            pole2 = lastvec.add(Vector(p2x, -p2y, 0))
                        else:
                            currentvec = Vector(x, -y, 0)
                            pole2 = Vector(p2x, -p2y, 0)

                        if not DraftVecUtils.equals(currentvec, lastvec):
                            # mainv = currentvec.sub(lastvec)
                            # pole1v = lastvec.add(pole1)
                            # pole2v = currentvec.add(pole2)
                            # print("cubic curve data:",
                            #       mainv.normalize(),
                            #       pole1v.normalize(),
                            #       pole2v.normalize())
                            _precision = 10**(-1*(2+Draft.precision()))
                            _d1 = pole1.distanceToLine(lastvec, currentvec)
                            _d2 = pole2.distanceToLine(lastvec, currentvec)
                            if True and \
                                    _d1 < _precision and \
                                    _d2 < _precision:
                                # print("straight segment")
                                _seg = Part.LineSegment(lastvec, currentvec)
                                seg = _seg.toShape()
                            else:
                                # print("cubic bezier segment")
                                b = Part.BezierCurve()
                                b.setPoles([lastvec, pole1, pole2, currentvec])
                                seg = b.toShape()
                            # print("connect ", lastvec, currentvec)
                            lastvec = currentvec
                            lastpole = ('cubic', pole2)
                            path.append(seg)
                elif (d == "Q" or d == "q") or (d == "T" or d == "t"):
                    smooth = (d == 'T' or d == 't')
                    if smooth:
                        piter = list(zip(pointlist[1::2],
                                         pointlist[1::2],
                                         pointlist[0::2],
                                         pointlist[1::2]))
                    else:
                        piter = list(zip(pointlist[0::4],
                                         pointlist[1::4],
                                         pointlist[2::4],
                                         pointlist[3::4]))
                    for px, py, x, y in piter:
                        if smooth:
                            if (lastpole is not None
                                    and lastpole[0] == 'quadratic'):
                                pole = lastvec.sub(lastpole[1]).add(lastvec)
                            else:
                                pole = lastvec
                        else:
                            if relative:
                                pole = lastvec.add(Vector(px, -py, 0))
                            else:
                                pole = Vector(px, -py, 0)
                        if relative:
                            currentvec = lastvec.add(Vector(x, -y, 0))
                        else:
                            currentvec = Vector(x, -y, 0)

                        if not DraftVecUtils.equals(currentvec, lastvec):
                            _precision = 20**(-1*(2+Draft.precision()))
                            _distance = pole.distanceToLine(lastvec,
                                                            currentvec)
                            if True and \
                                    _distance < _precision:
                                # print("straight segment")
                                _seg = Part.LineSegment(lastvec, currentvec)
                                seg = _seg.toShape()
                            else:
                                # print("quadratic bezier segment")
                                b = Part.BezierCurve()
                                b.setPoles([lastvec, pole, currentvec])
                                seg = b.toShape()
                            # print("connect ", lastvec, currentvec)
                            lastvec = currentvec
                            lastpole = ('quadratic', pole)
                            path.append(seg)
                elif (d == "Z") or (d == "z"):
                    if not DraftVecUtils.equals(lastvec, firstvec):
                        try:
                            seg = Part.LineSegment(lastvec, firstvec).toShape()
                        except Part.OCCError:
                            pass
                        else:
                            path.append(seg)
                    if path:
                        # The path should be closed by now
                        # sh = makewire(path, True)
                        sh = makewire(path, donttry=False)
                        if self.fill \
                                and len(sh.Wires) == 1 \
                                and sh.Wires[0].isClosed():
                            sh = Part.Face(sh)
                        sh = self.applyTrans(sh)
                        obj = self.doc.addObject("Part::Feature", pathname)
                        obj.Shape = sh
                        self.format(obj)
                        path = []
                        if firstvec:
                            # Move relative to recent draw command
                            lastvec = firstvec
                        point = []
                        # command = None
                        if self.currentsymbol:
                            self.symbols[self.currentsymbol].append(obj)
            if path:
                sh = makewire(path, checkclosed=False)
                # sh = Part.Wire(path)
                if self.fill and sh.isClosed():
                    sh = Part.Face(sh)
                sh = self.applyTrans(sh)
                obj = self.doc.addObject("Part::Feature", pathname)
                obj.Shape = sh
                self.format(obj)
                if self.currentsymbol:
                    self.symbols[self.currentsymbol].append(obj)
        # end process paths

        # Process rects
        if name == "rect":
            if not pathname:
                pathname = 'Rectangle'
            edges = []
            if "x" not in data:
                data["x"] = 0
            if "y" not in data:
                data["y"] = 0
            # Negative values are invalid
            _precision = 10**(-1*Draft.precision())
            if ('rx' not in data or data['rx'] < _precision) \
                    and ('ry' not in data or data['ry'] < _precision):
                # if True:
                p1 = Vector(data['x'],
                            -data['y'],
                            0)
                p2 = Vector(data['x'] + data['width'],
                            -data['y'],
                            0)
                p3 = Vector(data['x'] + data['width'],
                            -data['y'] - data['height'],
                            0)
                p4 = Vector(data['x'],
                            -data['y'] - data['height'],
                            0)
                edges.append(Part.LineSegment(p1, p2).toShape())
                edges.append(Part.LineSegment(p2, p3).toShape())
                edges.append(Part.LineSegment(p3, p4).toShape())
                edges.append(Part.LineSegment(p4, p1).toShape())
            else:
                # rounded edges
                rx = data.get('rx')
                ry = data.get('ry') or rx
                rx = rx or ry
                if rx > 2 * data['width']:
                    rx = data['width'] / 2.0
                if ry > 2 * data['height']:
                    ry = data['height'] / 2.0

                p1 = Vector(data['x'] + rx,
                            -data['y'] - data['height'] + ry,
                            0)
                p2 = Vector(data['x'] + data['width'] - rx,
                            -data['y'] - data['height'] + ry,
                            0)
                p3 = Vector(data['x'] + data['width'] - rx,
                            -data['y'] - ry,
                            0)
                p4 = Vector(data['x'] + rx,
                            -data['y'] - ry,
                            0)

                if rx >= ry:
                    e = Part.Ellipse(Vector(), rx, ry)
                    e1a = Part.Arc(e, math.radians(180), math.radians(270))
                    e2a = Part.Arc(e, math.radians(270), math.radians(360))
                    e3a = Part.Arc(e, math.radians(0), math.radians(90))
                    e4a = Part.Arc(e, math.radians(90), math.radians(180))
                    m = FreeCAD.Matrix()
                else:
                    e = Part.Ellipse(Vector(), ry, rx)
                    e1a = Part.Arc(e, math.radians(90), math.radians(180))
                    e2a = Part.Arc(e, math.radians(180), math.radians(270))
                    e3a = Part.Arc(e, math.radians(270), math.radians(360))
                    e4a = Part.Arc(e, math.radians(0), math.radians(90))
                    # rotate +90 degrees
                    m = FreeCAD.Matrix(0, -1, 0, 0, 1, 0)
                esh = []
                for arc, point in ((e1a, p1), (e2a, p2),
                                   (e3a, p3), (e4a, p4)):
                    m1 = FreeCAD.Matrix(m)
                    m1.move(point)
                    arc.transform(m1)
                    esh.append(arc.toShape())
                for esh1, esh2 in zip(esh[-1:] + esh[:-1], esh):
                    p1 = esh1.Vertexes[-1].Point
                    p2 = esh2.Vertexes[0].Point
                    if not DraftVecUtils.equals(p1, p2):
                        # straight segments
                        _sh = Part.LineSegment(p1, p2).toShape()
                        edges.append(_sh)
                    # elliptical segments
                    edges.append(esh2)
            sh = Part.Wire(edges)
            if self.fill:
                sh = Part.Face(sh)
            sh = self.applyTrans(sh)
            obj = self.doc.addObject("Part::Feature", pathname)
            obj.Shape = sh
            self.format(obj)
            if self.currentsymbol:
                self.symbols[self.currentsymbol].append(obj)

        # Process lines
        if name == "line":
            if not pathname:
                pathname = 'Line'
            p1 = Vector(data['x1'], -data['y1'], 0)
            p2 = Vector(data['x2'], -data['y2'], 0)
            sh = Part.LineSegment(p1, p2).toShape()
            sh = self.applyTrans(sh)
            obj = self.doc.addObject("Part::Feature", pathname)
            obj.Shape = sh
            self.format(obj)
            if self.currentsymbol:
                self.symbols[self.currentsymbol].append(obj)

        # Process polylines and polygons
        if name == "polyline" or name == "polygon":
            # A simpler implementation would be
            # _p = zip(points[0::2], points[1::2])
            # sh = Part.makePolygon([Vector(svgx,
            #                               -svgy,
            #                               0) for svgx, svgy in _p])
            #
            # but it would be more difficult to search for duplicate
            # points beforehand.
            if not pathname:
                pathname = 'Polyline'
            points = [float(d) for d in data['points']]
            FCC.PrintMessage('points %s\n' % str(points))
            lenpoints = len(points)
            if lenpoints >= 4 and lenpoints % 2 == 0:
                lastvec = Vector(points[0], -points[1], 0)
                path = []
                if name == 'polygon':
                    points = points + points[:2]  # emulate closepath
                for svgx, svgy in zip(points[2::2], points[3::2]):
                    currentvec = Vector(svgx, -svgy, 0)
                    if not DraftVecUtils.equals(lastvec, currentvec):
                        seg = Part.LineSegment(lastvec, currentvec).toShape()
                        # print("polyline seg ", lastvec, currentvec)
                        lastvec = currentvec
                        path.append(seg)
                if path:
                    sh = Part.Wire(path)
                    if self.fill and sh.isClosed():
                        sh = Part.Face(sh)
                    sh = self.applyTrans(sh)
                    obj = self.doc.addObject("Part::Feature", pathname)
                    obj.Shape = sh
                    if self.currentsymbol:
                        self.symbols[self.currentsymbol].append(obj)

        # Process ellipses
        if name == "ellipse":
            if not pathname:
                pathname = 'Ellipse'
            c = Vector(data.get('cx', 0), -data.get('cy', 0), 0)
            rx = data['rx']
            ry = data['ry']
            if rx > ry:
                sh = Part.Ellipse(c, rx, ry).toShape()
            else:
                sh = Part.Ellipse(c, ry, rx).toShape()
                m3 = FreeCAD.Matrix()
                m3.move(c)
                # 90
                rot90 = FreeCAD.Matrix(0, -1, 0, 0, 1, 0)
                m3 = m3.multiply(rot90)
                m3.move(c.multiply(-1))
                sh.transformShape(m3)
                # sh = sh.transformGeometry(m3)
            if self.fill:
                sh = Part.Wire([sh])
                sh = Part.Face(sh)
            sh = self.applyTrans(sh)
            obj = self.doc.addObject("Part::Feature", pathname)
            obj.Shape = sh
            self.format(obj)
            if self.currentsymbol:
                self.symbols[self.currentsymbol].append(obj)

        # Process circles
        if name == "circle" and "freecad:skip" not in data:
            if not pathname:
                pathname = 'Circle'
            c = Vector(data.get('cx', 0), -data.get('cy', 0), 0)
            r = data['r']
            sh = Part.makeCircle(r)
            if self.fill:
                sh = Part.Wire([sh])
                sh = Part.Face(sh)
            sh.translate(c)
            sh = self.applyTrans(sh)
            obj = self.doc.addObject("Part::Feature", pathname)
            obj.Shape = sh
            self.format(obj)
            if self.currentsymbol:
                self.symbols[self.currentsymbol].append(obj)

        # Process texts
        if name in ["text", "tspan"]:
            if "freecad:skip" not in data:
                FCC.PrintMessage("processing a text\n")
                if 'x' in data:
                    self.x = data['x']
                else:
                    self.x = 0
                if 'y' in data:
                    self.y = data['y']
                else:
                    self.y = 0
                if 'font-size' in data:
                    if data['font-size'] != 'none':
                        self.text = getsize(data['font-size'],
                                            'css' + str(self.svgdpi))
                else:
                    self.text = 1
            else:
                if self.lastdim:
                    _font_size = int(getsize(data['font-size']))
                    self.lastdim.ViewObject.FontSize = _font_size

        # Process symbols
        if name == "symbol":
            self.symbols[pathname] = []
            self.currentsymbol = pathname

        if name == "use":
            if "xlink:href" in data:
                symbol = data["xlink:href"][0][1:]
                if symbol in self.symbols:
                    FCC.PrintMessage("using symbol " + symbol + "\n")
                    shapes = []
                    for o in self.symbols[symbol]:
                        if o.isDerivedFrom("Part::Feature"):
                            shapes.append(o.Shape)
                    if shapes:
                        sh = Part.makeCompound(shapes)
                        v = Vector(float(data['x']), -float(data['y']), 0)
                        sh.translate(v)
                        sh = self.applyTrans(sh)
                        obj = self.doc.addObject("Part::Feature", symbol)
                        obj.Shape = sh
                        self.format(obj)
                else:
                    FCC.PrintMessage("no symbol data\n")

        FCC.PrintMessage("done processing element %d\n" % self.count)
    # startElement()

    def characters(self, content):
        """Read characters from the given string."""
        if self.text:
            FCC.PrintMessage("reading characters %s\n" % content)
            obj = self.doc.addObject("App::Annotation", 'Text')
            obj.LabelText = content.encode('latin1', 'ignore')  # use ignore to not break import if char is not found in latin1
            if self.currentsymbol:
                self.symbols[self.currentsymbol].append(obj)
            vec = Vector(self.x, -self.y, 0)
            if self.transform:
                vec = self.translateVec(vec, self.transform)
                # print("own transform: ", self.transform, vec)
            for transform in self.grouptransform[::-1]:
                # vec = self.translateVec(vec, transform)
                vec = transform.multiply(vec)
            # print("applying vector: ", vec)
            obj.Position = vec
            if FreeCAD.GuiUp:
                obj.ViewObject.FontSize = int(self.text)
                if self.fill:
                    obj.ViewObject.TextColor = self.fill
                else:
                    obj.ViewObject.TextColor = (0.0, 0.0, 0.0, 0.0)

    def endElement(self, name):
        """Finish processing the element indicated by the name.

        Parameters
        ----------
        name : str
            The name of the element
        """
        if name not in ["tspan"]:
            self.transform = None
            self.text = None
        if name == "g" or name == "svg":
            FCC.PrintMessage("closing group\n")
            self.grouptransform.pop()
        if name == "symbol":
            if self.doc.getObject("svgsymbols"):
                group = self.doc.getObject("svgsymbols")
            else:
                group = self.doc.addObject("App::DocumentObjectGroup",
                                           "svgsymbols")
            for o in self.symbols[self.currentsymbol]:
                group.addObject(o)
            self.currentsymbol = None

    def applyTrans(self, sh):
        """Apply transformation to the shape and return the new shape.

        Parameters
        ----------
        sh : Part.Shape or Draft.Dimension
            Object to be transformed
        """
        if isinstance(sh, Part.Shape):
            if self.transform:
                FCC.PrintMessage("applying object transform: %s\n"
                                 % self.transform)
                # sh = transformCopyShape(sh, self.transform)
                # see issue #2062
                sh = sh.transformGeometry(self.transform)
            for transform in self.grouptransform[::-1]:
                FCC.PrintMessage("applying group transform: %s\n" % transform)
                # sh = transformCopyShape(sh, transform)
                # see issue #2062
                sh = sh.transformGeometry(transform)
            return sh
        elif Draft.getType(sh) in ["Dimension","LinearDimension"]:
            pts = []
            for p in [sh.Start, sh.End, sh.Dimline]:
                cp = Vector(p)
                if self.transform:
                    FCC.PrintMessage("applying object transform: %s\n"
                                     % self.transform)
                    cp = self.transform.multiply(cp)
                for transform in self.grouptransform[::-1]:
                    FCC.PrintMessage("applying group transform: %s\n"
                                     % transform)
                    cp = transform.multiply(cp)
                pts.append(cp)
            sh.Start = pts[0]
            sh.End = pts[1]
            sh.Dimline = pts[2]

    def translateVec(self, vec, mat):
        """Translate (move) a point or vector by a matrix.

        Parameters
        ----------
        vec : Base::Vector3
            The original vector
        mat : Base::Matrix4D
            The translation matrix, from which only the elements 14, 24, 34
            are used.
        """
        v = Vector(mat.A14, mat.A24, mat.A34)
        return vec.add(v)

    def getMatrix(self, tr):
        """Return a FreeCAD matrix from an SVG transform attribute.

        Parameters
        ----------
        tr : str
            The type of transform: 'matrix', 'translate', 'scale',
            'rotate', 'skewX', 'skewY' and its value

        Returns
        -------
        Base::Matrix4D
            The translated matrix.
        """
        _op = '(matrix|translate|scale|rotate|skewX|skewY)'
        _val = '\((.*?)\)'
        _transf = _op + '\s*?' + _val
        transformre = re.compile(_transf, re.DOTALL)
        m = FreeCAD.Matrix()
        for transformation, arguments in transformre.findall(tr):
            _args_rep = arguments.replace(',', ' ').split()
            argsplit = [float(arg) for arg in _args_rep]
            # m.multiply(FreeCAD.Matrix(1, 0, 0, 0, 0, -1))
            # print('%s:%s %s %d' % (transformation, arguments,
            #                        argsplit, len(argsplit)))
            if transformation == 'translate':
                tx = argsplit[0]
                ty = argsplit[1] if len(argsplit) > 1 else 0.0
                m.move(Vector(tx, -ty, 0))
            elif transformation == 'scale':
                sx = argsplit[0]
                sy = argsplit[1] if len(argsplit) > 1 else sx
                m.scale(Vector(sx, sy, 1))
            elif transformation == 'rotate':
                cx = 0
                cy = 0
                angle = argsplit[0]
                if len(argsplit) >= 3:
                    cx = argsplit[1]
                    cy = argsplit[2]
                    m.move(Vector(cx, -cy, 0))
                # Mirroring one axis is equal to changing the direction
                # of rotation
                m.rotateZ(math.radians(-angle))
                if len(argsplit) >= 3:
                    m.move(Vector(-cx, cy, 0))
            elif transformation == 'skewX':
                _m = FreeCAD.Matrix(1,
                                    -math.tan(math.radians(argsplit[0])))
                m = m.multiply(_m)
            elif transformation == 'skewY':
                _m = FreeCAD.Matrix(1, 0, 0, 0,
                                    -math.tan(math.radians(argsplit[0])))
                m = m.multiply(_m)
            elif transformation == 'matrix':
                # transformation matrix:
                #    FreeCAD                 SVG
                # (+A -C +0 +E)           (A C 0 E)
                # (-B +D -0 -F)  = (-Y) * (B D 0 F) * (-Y)
                # (+0 -0 +1 +0)           (0 0 1 0)
                # (+0 -0 +0 +1)           (0 0 0 1)
                #
                # Put the first two rows of the matrix
                _m = FreeCAD.Matrix(argsplit[0], -argsplit[2],
                                    0, argsplit[4],
                                    -argsplit[1], argsplit[3],
                                    0, -argsplit[5])
                m = m.multiply(_m)
            # else:
            #    print('SKIPPED %s' % transformation)
            # print("m = ", m)
        # print("generating transformation: ", m)
        return m
    # getMatrix
# class svgHandler


def decodeName(name):
    """Decode encoded name.

    Parameters
    ----------
    name : str
        The string to decode.

    Returns
    -------
    tuple
    (string)
        A tuple containing the decoded `name` in 'utf8', otherwise in 'latin1'.
        If it fails it returns the original `name`.
    """
    try:
        decodedName = (name.decode("utf8"))
    except UnicodeDecodeError:
        try:
            decodedName = (name.decode("latin1"))
        except UnicodeDecodeError:
            FCC.PrintError("SVG error: "
                           "couldn't determine character encoding\n")
            decodedName = name
    return decodedName


def getContents(filename, tag, stringmode=False):
    """Get the contents of all occurrences of the given tag in the file.

    Parameters
    ----------
    filename : str
        A filename to scan for tags.
    tag : str
        An SVG tag to find inside a file, for example, `some`
        in <some id="12">information</some>
    stringmode : bool, optional
        The default is False.
        If False, `filename` is a path to a file.
        If True, `filename` is already a pointer to an open file.

    Returns
    -------
    dict
        A dictionary with tagids and the information associated with that id
        results[tagid] = information
    """
    result = {}
    if stringmode:
        contents = filename
    else:
        # Use the native Python open which was saved as `pythonopen`
        f = pythonopen(filename)
        contents = f.read()
        f.close()

    # Replace the newline character with a string
    # so that it's easiert to parse; later on the newline character
    # will be restored
    contents = contents.replace('\n', '_linebreak')
    searchpat = '<' + tag + '.*?</' + tag + '>'
    tags = re.findall(searchpat, contents)
    for t in tags:
        tagid = re.findall('id="(.*?)"', t)
        if tagid:
            tagid = tagid[0]
        else:
            tagid = 'none'
        res = t.replace('_linebreak', '\n')
        result[tagid] = res
    return result


def open(filename):
    """Open filename and parse using the svgHandler().

    Parameters
    ----------
    filename : str
        The path to the filename to be opened.

    Returns
    -------
    App::Document
        The new FreeCAD document object created, with the parsed information.
    """
    docname = os.path.split(filename)[1]
    doc = FreeCAD.newDocument(docname)
    doc.Label = docname[:-4]

    # Set up the parser
    parser = xml.sax.make_parser()
    parser.setFeature(xml.sax.handler.feature_external_ges, False)
    parser.setContentHandler(svgHandler())
    parser._cont_handler.doc = doc

    # Use the native Python open which was saved as `pythonopen`
    f = pythonopen(filename)
    parser.parse(f)
    f.close()
    doc.recompute()
    return doc


def insert(filename, docname):
    """Get an active document and parse using the svgHandler().

    If no document exist, it is created.

    Parameters
    ----------
    filename : str
        The path to the filename to be opened.
    docname : str
        The name of the active App::Document if one exists, or
        of the new one created.

    Returns
    -------
    App::Document
        The active FreeCAD document, or the document created if none exists,
        with the parsed information.
    """
    try:
        doc = FreeCAD.getDocument(docname)
    except NameError:
        doc = FreeCAD.newDocument(docname)
    FreeCAD.ActiveDocument = doc

    # Set up the parser
    parser = xml.sax.make_parser()
    parser.setFeature(xml.sax.handler.feature_external_ges, False)
    parser.setContentHandler(svgHandler())
    parser._cont_handler.doc = doc

    # Use the native Python open which was saved as `pythonopen`
    parser.parse(pythonopen(filename))
    doc.recompute()


def export(exportList, filename):
    """Export the SVG file with a given list of objects.

    The objects must be derived from Part::Feature, in order to be processed
    and exported.

    Parameters
    ----------
    exportList : list
        List of document objects to export.
    filename : str
        Path to the new file.

    Returns
    -------
    None
        If `exportList` doesn't have shapes to export.
    """
    _prefs = "User parameter:BaseApp/Preferences/Mod/Draft"
    svg_export_style = FreeCAD.ParamGet(_prefs).GetInt("svg_export_style")
    if svg_export_style != 0 and svg_export_style != 1:
        FCC.PrintMessage(translate("ImportSVG",
                                   "Unknown SVG export style, "
                                   "switching to Translated") + "\n")
        svg_export_style = 0

    # Determine the size of the page by adding the bounding boxes
    # of all shapes
    bb = None
    for ob in exportList:
        if ob.isDerivedFrom("Part::Feature"):
            if bb:
                bb.add(ob.Shape.BoundBox)
            else:
                bb = ob.Shape.BoundBox
    if bb:
        minx = bb.XMin
        maxx = bb.XMax
        miny = bb.YMin
        maxy = bb.YMax
    else:
        FCC.PrintError("The export list contains no shape\n")
        return

    if svg_export_style == 0:
        # translated-style exports get a bit of a margin
        margin = (maxx - minx) * 0.01
    else:
        # raw-style exports get no margin
        margin = 0

    minx -= margin
    maxx += margin
    miny -= margin
    maxy += margin
    sizex = maxx - minx
    sizey = maxy - miny
    miny += margin

    # Use the native Python open which was saved as `pythonopen`
    svg = pythonopen(filename, 'w')

    # Write header.
    # We specify the SVG width and height in FreeCAD's physical units (mm),
    # and specify the viewBox so that user units maps one-to-one to mm.
    svg.write('<?xml version="1.0"?>\n')
    svg.write('<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN"')
    svg.write(' "http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd">\n')
    svg.write('<svg')
    svg.write(' width="' + str(sizex) + 'mm" height="' + str(sizey) + 'mm"')
    if svg_export_style == 0:
        # translated-style exports have the viewbox starting at X=0, Y=0
        svg.write(' viewBox="0 0 ' + str(sizex) + ' ' + str(sizey) + '"')
    else:
        # Raw-style exports have the viewbox starting at X=xmin, Y=-ymax
        # We need the negative Y here because SVG is upside down, and we
        # flip the sketch right-way up with a scale later
        svg.write(' viewBox="%f %f %f %f"' % (minx, -maxy, sizex, sizey))
    svg.write(' xmlns="http://www.w3.org/2000/svg" version="1.1"')
    svg.write('>\n')

    # Write paths
    for ob in exportList:
        if svg_export_style == 0:
            # translated-style exports have the entire sketch translated
            # to fit in the X>0, Y>0 quadrant
            # svg.write('<g transform="translate('
            #           + str(-minx) + ',' + str(-miny + 2*margin)
            #           + ') scale(1,-1)">\n')
            svg.write('<g id="%s" transform="translate(%f,%f) '
                      'scale(1,-1)">\n' % (ob.Name, -minx, maxy))
        else:
            # raw-style exports do not translate the sketch
            svg.write('<g id="%s" transform="scale(1,-1)">\n' % ob.Name)
        svg.write(Draft.getSVG(ob))
        _label_enc = str(ob.Label.encode('utf8'))
        _label = _label_enc.replace('<', '&lt;').replace('>', '&gt;')
        # replace('"', "&quot;")
        svg.write('<title>%s</title>\n' % _label)
        svg.write('</g>\n')

    # Close the file
    svg.write('</svg>')
    svg.close()



"""Classes that define the viewproviders of custom scripted objects.

These classes define viewproviders for the custom objects
defined in `draftobjects`.
The viewproviders can be used only when the graphical interface
is available; in console mode the viewproviders are not available.
"""
